from __future__ import annotations

import ast
import collections.abc
import inspect
import pprint
import re
import typing
from copy import copy, deepcopy
from enum import Flag, auto

import numpy as np

from ..utils.logger import deprecated, logger
from ..utils.misc import serialize
from ..utils.tags import _not_found_, _undefined_
from ..utils.typing import array_type, isinstance_generic, is_int, is_tree


# fmt:off
class OpTags(Flag):
    # traversal operation 操作
    root  = auto()      # root node
    parent = auto()     # parent node
    children = auto()   # 所有子节点
    ancestors =auto()   # 所有祖先节点  
    descendants = auto()# 所有后代节点
    slibings=auto()     # 所有兄弟节点

    current = auto()    # current node
    next  = auto()      # next sibling
    prev  = auto()      # previous sibling

    # RESTful operation for CRUD
    find  = auto()      # GET
    update = auto()     # PUT
    insert = auto()     # POST
    remove = auto()     # DELETE


    append = auto()
    extend = auto()
    overwrite  = auto()    # 强制覆盖
    
 

    call  = auto()      # call function
    exists = auto()
    is_leaf = auto()
    is_list=auto()
    is_dict=auto()
    check_type = auto() # check type
    keys=auto()         # 返回键值
    search = auto()     # search by query return idx
    dump  = auto()      # rescurive get all data

    # for sequence
    reduce = auto()
    sort  = auto()

    # predicate 谓词
    check  = auto()
    count  = auto()

    # boolean
    equal  = auto()
    le   = auto()
    ge   = auto()
    less  = auto()
    greater = auto()
# fmt:on


QueryLike = dict | OpTags | None


class Query:
    def __init__(self, query: QueryLike = None, only_first=True, **kwargs) -> None:
        if query is _not_found_ or query is None:
            query = {}
        if isinstance(query, Query):
            self._query = Query._parser(query._query, **kwargs)
        else:
            self._query = Query._parser(query, **kwargs)

        self._only_first = only_first

    def __str__(self) -> str:
        p = self._query

        if not isinstance(p, dict):
            return str(p)
        else:
            m_str = ",".join([f"{k}:{Path._to_str(v)}" for k, v in p.items()])
            return f"?{{{m_str}}}"

    @classmethod
    def _parser(cls, query: QueryLike, **kwargs) -> dict:
        if query is None:
            query = {".": Path._op_fetch}

        elif isinstance(query, Path.tags):
            query = {".": f"${query.name}"}

        elif isinstance(query, str) and query.startswith("$"):
            query = {".": query}

        elif isinstance(query, str):
            query = {f"@{Path.id_tag_name}": query}

        elif isinstance(query, dict):
            query = {k: Query._parser(v) for k, v in query.items()}

        else:
            raise TypeError(f"{(query)}")

        if isinstance(query, dict):
            return update_tree(query, kwargs)

        else:
            return query

    def __call__(self, target, *args, **kwargs) -> typing.Any:
        return self.check(target, *args, **kwargs)

    @staticmethod
    def _eval_one(target, k, v) -> typing.Any:
        res = False
        if k == "." or k is OpTags.current:
            res = Query._q_equal(target, v)

        elif isinstance(k, str) and k.startswith("@"):
            if hasattr(target.__class__, k[1:]):
                res = Query._q_equal(getattr(target, k[1:], _not_found_), v)

            elif isinstance(target, collections.abc.Mapping):
                res = Query._q_equal(target.get(k, _not_found_), v)

        return res

    def _eval(self, target) -> bool:
        return all([Query._eval_one(target, k, v) for k, v in self._query.items()])

    def check(self, target) -> bool:
        res = self._eval(target)

        if isinstance(res, list):
            return all(res)
        else:
            return bool(res)

    def find_next(self, target, start: int | None, **kwargs) -> typing.Tuple[typing.Any, int | None]:
        next = start
        return _not_found_, next

    @staticmethod
    def _q_equal(target, value) -> bool:
        if isinstance(target, collections.abc.Sequence):
            return value in target
        else:
            return target == value

    # fmt: off
    _q_neg         =np.negative   
    _q_add         =np.add     
    _q_sub         =np.subtract   
    _q_mul         =np.multiply   
    _q_matmul      =np.matmul    
    _q_truediv     =np.true_divide 
    _q_pow         =np.power    
    _q_equal       =np.equal    
    _q_ne          =np.not_equal  
    _q_lt          =np.less     
    _q_le          =np.less_equal  
    _q_gt          =np.greater   
    _q_ge          =np.greater_equal
    _q_radd        =np.add     
    _q_rsub        =np.subtract   
    _q_rmul        =np.multiply   
    _q_rmatmul     =np.matmul    
    _q_rtruediv    =np.divide    
    _q_rpow        =np.power    
    _q_abs         =np.abs     
    _q_pos         =np.positive   
    _q_invert      =np.invert    
    _q_and         =np.bitwise_and 
    _q_or          =np.bitwise_or  
    _q_xor         =np.bitwise_xor 
    _q_rand        =np.bitwise_and 
    _q_ror         =np.bitwise_or  
    _q_rxor        =np.bitwise_xor 
    _q_rshift      =np.right_shift 
    _q_lshift      =np.left_shift  
    _q_rrshift     =np.right_shift 
    _q_rlshift     =np.left_shift  
    _q_mod         =np.mod     
    _q_rmod        =np.mod     
    _q_floordiv    =np.floor_divide 
    _q_rfloordiv_  =np.floor_divide 
    _q_trunc       =np.trunc    
    _q_round       =np.round    
    _q_floor       =np.floor    
    _q_ceil        =np.ceil     
    # fmt: on


def as_query(query: QueryLike = None, **kwargs) -> Query | slice:
    if isinstance(query, slice):
        return query
    elif isinstance(query, Query):
        return query
    else:
        return Query(query, **kwargs)


class PathError(Exception):
    def __init__(self, path: typing.List[PathLike], message: str | None = None) -> None:
        if message is None:
            message = f"PathError: {Path(path)}"
        else:
            message = f"PathError: {Path(path)}: {message}"
        super().__init__(message)


class Path(list):
    """Path用于描述数据的路径, 在 HTree ( Hierarchical Tree) 中定位Element, 其语法是 JSONPath 和 XPath的变体，
    并扩展谓词（predicate）语法/查询选择器。

    HTree:
        Hierarchical Tree 半结构化树状数据，树节点具有 list或dict类型，叶节点为 list和dict 之外的primary数据类型，
    包括 int，float,string 和 ndarray。

    基本原则是用python 原生数据类型（例如，list, dict,set,tuple）等

    DELIMITER=`/` or `.`

    | Python 算符          | 字符形式          | 描述
    | ----             |---            | ---
    | N/A              | `$`            | 根对象 （ TODO：Not Implemented ）
    | None             | `@`            | 空选择符，当前对象。当以Path以None为最后一个item时，表示所指元素为leaf节点。
    | `__truediv__`,`__getattr___` | DELIMITER (`/` or `.`)  | 子元素选择符, DELIMITER 可选
    | `__getitem__`         | `[index|slice|selector]`| 数组元素选择符，index为整数,slice，或selector选择器（predicate谓词）

    predicate  谓词, 过滤表达式，用于过滤数组元素.

    | `set`             | `[{a,b,1}]`        | 返回dict, named并集运算符，用于组合多个子元素选择器，并将element作为返回的key， {'a':@[a], 'b':@['b'], 1:@[1] }
    | `list`            | `["a",b,1]`        | 返回list, 并集运算符，用于组合多个子元素选择器，[@[a], @['b'], @[1]]
    | `slice`            | `[start:end:step]`，   | 数组切片运算符, 当前元素为 ndarray 时返回数组切片 @[<slice>]，当前元素为 dict,list 以slice选取返回 list （generator），
    | `slice(None) `        | `*`            | 通配符，匹配任意字段或数组元素，代表所有子节点（children）
    |                | `..`           | 递归下降运算符 (Not Implemented)
    | `dict` `{$eq:4, }`      | `[?(expression)]`     | 谓词（predicate）或过滤表达式，用于过滤数组元素.
    |                | `==、!=、<、<=、>、>=`   | 比较运算符

    Examples

    | Path               | Description
    | ----               | ---
    | `a/b/c`              | 选择a节点的b节点的c节点
    | `a/b/c/1`             | 选择a节点的b节点的c节点的第二个元素
    | `a/b/c[1:3]`           | 选择a节点的b节点的c节点的第二个和第三个元素
    | `a/b/c[1:3:2]`          | 选择a节点的b节点的c节点的第二个和第三个元素
    | `a/b/c[1:3:-1]`          | 选择a节点的b节点的c节点的第三个和第二个元素
    | `a/b/c[d,e,f]`          |
    | `a/b/c[{d,e,f}]          |
    | `a/b/c[{value:{$le:10}}]/value  |
    | `a/b/c.$next/           |

    """

    id_tag_name = "name"
    delimiter = "/"
    tags = OpTags

    def __init__(self, *args, **kwargs):
        super().__init__(args[0] if len(args) == 1 and isinstance(args[0], list) else Path._parser(args), **kwargs)

    def __repr__(self):
        return Path._to_str(self)

    def __str__(self):
        return Path._to_str(self)

    def __hash__(self) -> int:
        return self.__str__().__hash__()

    def __copy__(self) -> Path:
        return self.__class__(deepcopy(self[:]))

    def as_url(self) -> str:
        return Path._to_str(self)

    @property
    def is_leaf(self) -> bool:
        return len(self) > 0 and self[-1] is None

    @property
    def is_root(self) -> bool:
        return len(self) == 0

    @property
    def is_regular(self) -> bool:
        return not self.is_generator

    @property
    def is_generator(self) -> bool:
        return any([isinstance(v, (slice, dict)) for v in self])

    @property
    def parent(self) -> Path:
        if len(self) == 0:
            logger.warning("Root node hasn't parents")
            return self
        else:
            return Path(self[:-1])

    @property
    def children(self) -> Path:
        if self.is_leaf:
            raise RuntimeError("Leaf node hasn't child!")
        other = copy(self)
        other.append(slice(None))
        return other

    @property
    def slibings(self):
        return self._parent.children

    @property
    def next(self) -> Path:
        other = copy(self)
        other.append(Path.tags.next)
        return other

    def prepend(self, d) -> Path:
        res = as_path(d)
        return res.append(self)

    def append(self, d) -> Path:
        return Path._resolve(Path._parser_iter(d), self)

    def extend(self, d: list) -> Path:
        return Path._resolve(d, self)

    def with_suffix(self, pth: str) -> Path:
        pth = Path(pth)
        if len(self) == 0:
            return pth
        else:
            res = copy(self)
            if isinstance(res[-1], str) and isinstance(pth[0], str):
                res[-1] += pth[0]
                res.extend(pth[1:])
            else:
                res.extend(pth[:])
        return res

    def __truediv__(self, p) -> Path:
        return copy(self).append(p)

    def __add__(self, p) -> Path:
        return copy(self).append(p)

    def __iadd__(self, p) -> Path:
        return self.append(p)

    def __eq__(self, other) -> bool:
        if isinstance(other, list):
            return super().__eq__(other)
        elif isinstance(other, Path):
            return super().__eq__(other[:])
        else:
            return False

    def collapse(self, idx=None) -> Path:
        """
        - 从路径中删除非字符元素，例如 slice, dict, set, tuple，int。用于从 default_value 中提取数据
        - 从路径中删除指定位置idx: 的元素

        """
        if idx is None:
            return Path([p for p in self if isinstance(p, str)])
        else:
            return Path(self[:idx] + self[idx + 1 :])

    @staticmethod
    def reduce(path: list) -> list:
        if len(path) < 2:
            return path
        elif isinstance(path[0], set) and path[1] in path[0]:
            return Path.reduce(path[1:])
        elif isinstance(path[0], slice) and isinstance(path[1], int):
            start = path[0].start if path[0].start is not None else 0
            step = path[0].step if path[0].step is not None else 1
            stop = start + step * path[1]
            if path[0].stop is not None and stop > path[0].stop:
                raise IndexError(f"index {stop} is out of range")
            return [stop, *Path.reduce(path[2:])]
        else:
            return path

    @staticmethod
    def normalize(p: typing.Any, raw=False) -> typing.Any:
        if p is None:
            res = []
        elif isinstance(p, Path):
            res = p[:]
        elif isinstance(p, str):
            res = Path._parser_str(p)
        elif isinstance(p, (int, slice)):
            res = p
        elif isinstance(p, list):
            res = sum((([v] if not isinstance(v, list) else v) for v in map(Path.normalize, p)), list())
        elif isinstance(p, tuple):
            if len(p) == 1:
                res = Path.normalize(p[0])
            else:
                res = tuple(map(Path.normalize, p))
        elif isinstance(p, collections.abc.Set):
            res = set(map(Path.normalize, p))
        elif isinstance(p, collections.abc.Mapping):
            res = {Path.normalize(k): Path.normalize(v, raw=True) for k, v in p.items()}
        else:
            res = p
            # raise TypeError(f"Path.normalize() only support str or Path, not {type(p)}")

        # if not raw and not isinstance(res, list):
        #   res = [res]

        return res

    @staticmethod
    def _resolve(source, target: list | None = None) -> list:
        """Make the path absolute, resolving all Path.tags (i.e. tag.root, tags.parent)"""
        if target is None:
            target = []

        for p in source:
            if p is Path.tags.parent:
                if len(target) > 0:
                    target.pop()
            elif p is Path.tags.root:
                target.clear()
                list.append(target, Path.tags.root)
            else:
                list.append(target, p)

        return target

    @staticmethod
    def _expand(target: typing.Any):
        if isinstance(target, collections.abc.Generator):
            res = [Path._expand(v) for v in target]
            if len(res) > 1 and isinstance(res[0], tuple) and len(res[0]) == 2:
                res = dict(*res)
            elif len(res) == 1:
                res = res[0]
        else:
            res = target

        return res

    @staticmethod
    def _to_str(p: typing.Any) -> str:
        if isinstance(p, list):
            return Path.delimiter.join([Path._to_str(s) for s in p])
        elif isinstance(p, str):
            return p
        elif isinstance(p, int):
            return str(p)
        elif isinstance(p, tuple):
            m_str = ",".join([Path._to_str(s) for s in p])
            return f"({m_str})"
        elif isinstance(p, set):
            m_str = ",".join([Path._to_str(s) for s in p])
            return f"{{{m_str}}}"
        elif isinstance(p, slice):
            if p.start is None and p.stop is None and p.step is None:
                return "*"
            else:
                return f"{p.start}:{p.stop}:{p.step}"
        elif p is None:
            return ""
        else:
            return str(p)
            # raise NotImplementedError(f"Not support Query,list,mapping,tuple to str,yet! {(p)}")

    # example:
    # a/b_c6/c[{value:{$le:10}}][value]/D/[1，2/3，4，5]/6/7.9.8

    PATH_PATTERN = re.compile(r"(?P<key>[^\[\]\/\,]+)(\[(?P<selector>[^\[\]]+)\])?")

    # 正则表达式解析，匹配一段被 {} 包裹的字符串
    PATH_REGEX_DICT = re.compile(r"\{(?P<selector>[^\{\}]+)\}")

    @staticmethod
    def _parser_selector(s: str | list) -> PathLike:
        if isinstance(s, str):
            s = s.strip(" ")

        if not isinstance(s, str):
            item = s
        elif s.startswith(("[", "(", "{")) and s.endswith(("}", ")", "]")):
            tmp = ast.literal_eval(s)
            if isinstance(tmp, dict):
                item = Query(tmp)  # {Path._parser_str_one(k): d for k, d in tmp.items()}
            elif isinstance(tmp, set):
                item = set([Path._parser_selector(k) for k in tmp])
            elif isinstance(tmp, tuple):
                item = tuple([Path._parser_selector(k) for k in tmp])
            elif isinstance(tmp, list):
                item = [Path._parser_selector(k) for k in tmp]

        elif s.startswith("(") and s.endswith(")"):
            tmp: dict = ast.literal_eval(s)
            item = {Path._parser_selector(k): d for k, d in tmp.items()}
        elif ":" in s:
            tmp = s.split(":")
            if len(tmp) == 2:
                item = slice(int(tmp[0]), int(tmp[1]))
            elif len(tmp) == 3:
                item = slice(int(tmp[0]), int(tmp[1]), int(tmp[2]))
            else:
                raise ValueError(f"Invalid slice {s}")
        elif s == "*":
            item = slice(None)
        elif s == "..":
            item = Path.tags.parent
        elif s == "...":
            item = Path.tags.ancestors
        elif s == "**":
            item = Path.tags.descendants
        elif s == ".":
            item = Path.tags.current
        elif s.isnumeric():
            item = int(s)
        elif s.startswith("$") and hasattr(Path.tags, s[1:]):
            item = Path.tags[s[1:]]
        else:
            item = s

        return item

    @staticmethod
    def _parser_iter(path: typing.Any) -> typing.Generator[PathLike, None, None]:
        if isinstance(path, str):
            if path.startswith("/"):
                yield Path.tags.root
                path = path[1:]
            elif path.isidentifier():
                yield path
                return

            for match in Path.PATH_PATTERN.finditer(path):
                key = match.group("key")

                if key is None:
                    pass

                elif (tmp := is_int(key)) is not False:
                    yield tmp

                elif key == "*":
                    yield Path.tags.children

                elif key == "..":
                    yield Path.tags.parent

                elif key == "...":
                    yield Path.tags.ancestors

                else:
                    yield key

                selector = match.group("selector")
                if selector is not None:
                    yield Path._parser_selector(selector)

        elif isinstance(path, Path.tags):
            yield path
            
        elif isinstance(path, (int, slice, set)):
            yield path

        elif isinstance(path, collections.abc.Sequence):
            for item in path:
                yield from Path._parser_iter(item)

        else:
            yield Query(path)

    @staticmethod
    def _parser(path) -> list:
        """Parse the PathLike to list"""
        return [*Path._parser_iter(path)]

    ###########################################################
    # RESTful API:

    # 非幂等
    def insert(self, target: typing.Any, value, *args, **kwargs) -> typing.Tuple[typing.Any, Path]:
        """
        根据路径（self）向 target 添加元素。
        当路径指向位置为空时，创建（create）元素
        当路径指向位置为 list 时，追加（ insert ）元素
        当路径指向位置为非 list 时，合并为 [old,new]
        当路径指向位置为 dict, 添加值亦为 dict 时，根据 key 递归执行 insert

        返回新添加元素的路径

        对应 RESTful 中的 post，非幂等操作
        """
        return Path._do_update(target, self[:], value, Path.tags.insert, *args, **kwargs)

    # 幂等
    def update(self, target: typing.Any, *args, **kwargs) -> typing.Any:
        """
        根据路径（self）更新 target 中的元素。
        当路径指向位置为空时，创建（create）元素
        当路径指向位置为 dict, 添加值亦为 dict 时，根据 key 递归执行 update
        当路径指向位置为空时，用新的值替代（replace）元素

        对应 RESTful 中的 put， 幂等操作

        返回修改后的target
        """
        return Path._do_update(target, self[:], *args, **kwargs)

    def remove(self, target: typing.Any, *args, **kwargs) -> typing.Tuple[typing.Any, int]:
        """根据路径（self）删除 target 中的元素。

        if quiet is False then raise KeyError if the path is not found

        对应 RESTful 中的 delete， 幂等操作

        返回修改后的target和删除的元素的个数
        """
        return Path._do_update(target, self[:], _not_found_, Path.tags.overwrite, *args, **kwargs)

    def find(self, source: typing.Any, *args, **kwargs) -> typing.Any:
        """
        根据路径（self）查询元素。只读，不会修改 target
        对应 RESTful 中的 read，幂等操作
        """
        return Path._do_find(source, self[:], *args, **kwargs)

    def for_each(self, source, *args, **kwargs) -> typing.Generator[typing.Tuple[int | str, typing.Any], None, None]:
        yield from Path._do_for_each(source, self[:], *args, **kwargs)

    def search(self, source, query, *args, **kwargs) -> int | None:
        return Path._do_find(source, self[:], Path.tags.search, query, *args, **kwargs)

    def keys(self, source) -> typing.Generator[str, None, None]:
        yield from Path._do_for_each(source, Path.tags.keys)

    def copy(self, source: typing.Any, target: typing.Any = None, *args, **kwargs):
        """copy object to target"""
        return self._do_copy(source, target, self[:], *args, **kwargs)

    # --------------------------------------------------------------------------------------
    # 以下为简写，等价于上述方法，不应被重载

    def put(self, target: typing.Any, value, _op=None) -> typing.Any:
        return self.update(target, value, _op=_op)

    def get(self, source: typing.Any, default_value=_not_found_):
        return self.find(source, default_value=default_value)

    # End API
    ###########################################################

    @staticmethod
    def _apply_op(op: Path.tags, obj: typing.Any, *args, **kwargs):
        if op is Path.tags.find or op is None:
            return obj

        elif isinstance(op, Path.tags):
            func = getattr(Path, f"_op_{op.name}", None)
        else:
            func = op

        if not callable(func):
            raise RuntimeError(f"Unknown operator {op}")

        try:
            res = func(obj, *args, **kwargs)
        except Exception as error:
            raise RuntimeError(f'Fail to call "{op}"! {func}') from error

        return res

    @staticmethod
    def _do_copy(source: typing.Any, target: typing.Any, *args, **kwargs) -> typing.Any:
        raise NotImplementedError(f"TODO")

    @staticmethod
    def _update_tree(target, *sources, _idempotent=True, **kwargs):
        if len(kwargs) > 0:
            sources = [*sources, kwargs]

        for source in sources:
            if source is _not_found_:  # 不更新
                continue
            elif (target is _not_found_ or target is None) and not isinstance(source, dict):  # 直接替换
                target = source
            elif isinstance(source, dict):  # 合并 dict
                # if not isinstance(target, dict) and hasattr(target.__class__, "update"):
                #     # target is object with update method
                #     target.update(source)
                # else:
                if target is _not_found_ or target is None:
                    target = {}

                for key, value in source.items():
                    Path(key).update(target, value, _idempotent=_idempotent)

            elif _idempotent is True:  # 幂等操作，直接替换对象 target
                target = source

            elif isinstance(target, list):  # 合并 sequence
                if isinstance(source, list):
                    target.extend(source)
                else:
                    target.append(source)
            elif isinstance(source, list):
                target = [target, *source]
            else:
                target = [target, source]

        return target

    @staticmethod
    def _do_update(
        target: typing.Any,
        path: typing.List | None,
        value: typing.Any = _undefined_,
        _op: OpTags | None = None,
        *args,
        **kwargs,
    ) -> typing.Any:
        if value is _undefined_ and _op is None:
            return target
        if path is None:
            path = []
        elif not isinstance(path, list):
            path = [path]

        if len(path) == 0 and target is value:
            pass

        elif len(path) == 0:
            if target is value:
                pass

            elif value is _not_found_ and _op is None:
                pass

            elif _op is Path.tags.overwrite or target is _not_found_:
                target = value

            elif hasattr(target.__class__, "_update_"):
                target._update_([], value, _op, *args, **kwargs)

            elif isinstance(value, dict) and (_op is None or _op is Path.tags.update):
                for k, v in value.items():
                    target = Path._do_update(target, as_path(k)[:], v, _op, *args, **kwargs)

            elif _op is None:
                target = value

            elif _op is Path.tags.insert or _op is Path.tags.append:
                if isinstance(target, collections.abc.MutableSequence) and not isinstance(target, str):
                    target.append(value)
                else:
                    target = [target, value]

            elif _op is Path.tags.extend:
                if target is _not_found_:
                    target = value

                elif (
                    isinstance(target, collections.abc.MutableSequence)
                    and not isinstance(target, str)
                    and isinstance(value, list)
                ):
                    target.extend(value)
                else:
                    target = [target, value]

            else:
                raise NotImplementedError(f" {_op} {target} {value}")
                target = Path._apply_op(_op, target, value, *args, **kwargs)

        else:
            if target is _not_found_:
                if isinstance(path[0], str):
                    target = {}
                else:
                    target = []

            obj = target

            path_length = len(path)

            idx = 0

            while idx < path_length:
                key = path[idx]

                if not is_tree(obj):
                    raise TypeError(f" {obj} is not a Tree! key= {key}")

                elif key is Path.tags.current or key is None:
                    pass

                elif key is Path.tags.parent:
                    obj = getattr(obj, "_parent", _not_found_)

                elif key is Path.tags.children:
                    if hasattr(obj.__class__, "for_each"):
                        for k, v in enumerate(obj.for_each()):
                            old_value = v
                            new_value = Path._do_update(old_value, path[idx + 1 :], _op=_op, **kwargs)
                            if new_value is not old_value:
                                obj[k] = new_value

                    elif isinstance(obj, collections.abc.MutableMapping):
                        for k, v in obj.items():
                            old_value = v
                            new_value = Path._do_update(old_value, path[idx + 1 :], _op=_op, **kwargs)
                            if new_value is not old_value:
                                obj[k] = new_value

                    elif isinstance(obj, collections.abc.MutableSequence):
                        for k, v in enumerate(obj):
                            old_value = v
                            new_value = Path._do_update(old_value, path[idx + 1 :], _op=_op, **kwargs)
                            if new_value is not old_value:
                                obj[k] = new_value

                    else:
                        raise KeyError(f"{type(obj)} has not children!")

                    break

                elif isinstance(key, Path.tags):
                    raise NotImplementedError(key)

                elif hasattr(obj.__class__, "_update_"):
                    # FIXME：更改好的实现，但需要 debug
                    if idx == path_length - 1:
                        obj._update_(key, value, _op, *args, **kwargs)
                    else:
                        obj._update_(key, _not_found_, _idempotent=_op)

                    obj = obj._find_(key, default_value=_not_found_)
                    # old_value = obj._find_(key, default_value=_not_found_)

                    # if is_tree(old_value) and idx < path_length - 1:
                    #     new_value = old_value
                    # elif idx == path_length - 1:
                    #     new_value = Path._do_update(old_value, [], value, _idempotent=_idempotent, **kwargs)
                    # else:
                    #     new_value = Path._do_update(old_value, path[idx + 1], _not_found_, _idempotent=_idempotent)
                    #     idx -= 1

                    # if old_value is not new_value:
                    #     obj = obj._update_(key, new_value)
                    # else:
                    #     obj = new_value

                elif isinstance(key, str) and key.isidentifier() and hasattr(obj.__class__, key):
                    old_value = getattr(obj, key, _not_found_)

                    if is_tree(old_value) and idx < path_length - 1:
                        new_value = old_value
                    elif idx == path_length - 1:
                        new_value = Path._do_update(old_value, [], value, _op=_op, **kwargs)
                    else:
                        new_value = Path._do_update(old_value, path[idx + 1], _not_found_, _op=_op)
                        idx -= 1

                    if old_value is _not_found_ or old_value is not new_value:
                        setattr(obj, key, new_value)
                        obj = getattr(obj, key)
                    else:
                        obj = new_value

                elif isinstance(obj, collections.abc.MutableMapping):
                    old_value = obj.get(key, _not_found_)

                    if is_tree(old_value) and idx < path_length - 1:
                        new_value = old_value
                    elif idx == path_length - 1:
                        new_value = Path._do_update(old_value, [], value, _op=_op, *args, **kwargs)
                    else:
                        new_value = Path._do_update(old_value, path[idx + 1], _not_found_, _op=_op)

                    if old_value is _not_found_ or old_value is not new_value:
                        obj[key] = new_value
                        obj = obj[key]
                    else:
                        obj = new_value

                elif isinstance(obj, collections.abc.MutableSequence) and isinstance(key, (Path.tags, slice)):
                    for v in obj[key]:
                        Path._do_update(v, path[idx + 1 :], value, **kwargs)
                    break

                elif isinstance(obj, collections.abc.MutableSequence):
                    if key is None:
                        key = len(obj)

                    if isinstance(key, int):
                        if key >= (len(obj)):
                            obj.extend([_not_found_] * (key - len(obj) + 1))

                        old_value = obj[key]

                        if is_tree(old_value) and idx < path_length - 1:
                            new_value = old_value
                        elif idx == path_length - 1:
                            new_value = Path._do_update(old_value, [], value, _op=_op, **kwargs)
                        else:
                            new_value = Path._do_update(old_value, path[idx + 1], _not_found_, _op=_op)

                    elif isinstance(key, str):
                        query = Query(key)
                        for jdx, old_value in enumerate(obj):
                            if query.check(old_value):
                                path[idx] = key = jdx
                                break
                        else:
                            old_value = {f"@{Path.id_tag_name}": key}
                            path[idx] = key = len(obj)
                            obj.append(old_value)
                            _op = None

                        if is_tree(old_value) and idx < path_length - 1:
                            new_value = old_value
                        elif idx == path_length - 1:
                            new_value = Path._do_update(old_value, [], value, _op=_op, **kwargs)
                        else:
                            raise ValueError(f"Unknown value {old_value}")

                    elif isinstance(obj, collections.abc.MutableSequence) and isinstance(key, Query):
                        query = key
                        for key, old_value in enumerate(obj):
                            if query.check(old_value):
                                path[idx] = key
                                break
                        else:
                            raise KeyError(f"Can not update {obj} at {key}!")

                        if is_tree(old_value) and idx < path_length - 1:
                            new_value = old_value
                        elif idx == path_length - 1:
                            new_value = Path._do_update(old_value, [], value, _op=_op, **kwargs)
                        elif old_value is _not_found_:
                            new_value = Path._do_update(
                                {f"@{Path.id_tag_name}": key}, path[idx + 1], _not_found_, _op=_op
                            )
                        else:
                            raise ValueError(f"Unknown value {old_value}")

                    if old_value is _not_found_ or old_value is not new_value:
                        obj[key] = new_value
                        obj = obj[key]

                    else:
                        obj = new_value

                else:
                    raise KeyError(f"Can not find {key} in {obj}!")

                idx += 1

        return target

    @staticmethod
    def _do_find(source: typing.Any, path: list, _op=None, *args, default_value=_not_found_, **kwargs) -> typing.Any:
        if path is None:
            path = []
        elif not isinstance(path, list):
            path = [path]

        obj = source

        for idx, key in enumerate(path):
            if key is None or key is Path.tags.current or obj is _not_found_ or obj is None:
                pass

            elif (
                isinstance(key, str)
                and key.isidentifier()
                and (attr := getattr(obj, key, _not_found_)) is not _not_found_
            ):
                obj = attr

            elif len(path) > 1 and path[1] is Path.tags.next:
                if isinstance(key, int):
                    new_key = key + 1
                else:
                    raise NotImplementedError(f"{type(obj)} {path}")

                res = Path._do_find(obj, [new_key] + path[idx + 2 :], *args, **kwargs)

            elif len(path) > 1 and path[1] is Path.tags.prev:
                if isinstance(key, int):
                    new_key = key - 1
                else:
                    raise NotImplementedError(f"{type(obj)} {path}")

                res = Path._do_find(obj, [new_key] + path[idx + 2 :], *args, **kwargs)

            elif key is Path.tags.parent:
                obj = getattr(obj, "_parent", _not_found_)

            elif key is Path.tags.ancestors:
                # 逐级查找上层 _parent, 直到找到
                while obj is not None and obj is not _not_found_:
                    if not isinstance(obj, collections.abc.Sequence):
                        tmp = Path._do_find(obj, path[idx + 1 :], *args, default_value=_not_found_, **kwargs)
                        if tmp is not _not_found_:
                            obj=tmp
                            break
                    obj = getattr(obj, "_parent", _not_found_)
                    # Path._do_find(obj, [Path.tags.parent], default_value=_not_found_)
                else:
                    obj = _not_found_

                break

            elif key is Path.tags.descendants:
                # 遍历访问所有叶节点
                if isinstance(obj, collections.abc.Mapping):
                    obj = {
                        k: Path._do_find(v, [Path.tags.descendants] + path[idx + 1 :], *args, **kwargs)
                        for k, v in obj.items()
                    }

                elif isinstance(obj, collections.abc.Iterable):
                    obj = [Path._do_find(v, [Path.tags.descendants] + path[idx + 1 :], *args, **kwargs) for v in obj]

                elif len(path) > 0:
                    obj = Path._do_find(obj, path[idx + 1 :], *args, **kwargs)

            elif key is Path.tags.children:
                if isinstance(obj, collections.abc.Mapping):
                    obj = {k: Path._do_find(v, path[idx + 1 :], *args, **kwargs) for k, v in obj.items()}

                elif isinstance(obj, collections.abc.Iterable):
                    obj = [Path._do_find(v, path[idx + 1 :], *args, **kwargs) for v in obj]

                else:
                    obj = _not_found_
                break

            elif key is Path.tags.slibings:
                parent = getattr(obj, "_parent", _not_found_)

                if isinstance(parent, collections.abc.Mapping):
                    obj = {
                        k: Path._do_find(v, path[idx + 1 :], *args, **kwargs) for k, v in parent.items() if v is not obj
                    }

                elif isinstance(parent, collections.abc.Iterable):
                    obj = [Path._do_find(v, path[idx + 1 :], *args, **kwargs) for v in parent if v is not obj]

                else:
                    obj = []
                break
            elif isinstance(key, set):
                obj = {k: Path._do_find(obj, as_path(k)[:] + path[idx + 1 :], *args, **kwargs) for k in key}
                break
            elif hasattr(obj.__class__, "_find_"):
                obj = obj._find_(key, default_value=_not_found_)

            elif isinstance(obj, collections.abc.Mapping):
                obj = obj.get(key, _not_found_)

            elif isinstance(obj, collections.abc.Sequence) and not isinstance(obj, str):
                if isinstance(key, int):
                    if key < len(obj):
                        obj = obj[key]
                    else:
                        obj = _not_found_
                elif isinstance(key, slice):
                    obj = [Path._do_find(s, path[idx + 1 :], *args, **kwargs) for s in obj[key]]
                    break

                else:
                    idx, value = Path._op_search(obj, key)

                    if idx is None:
                        obj = _not_found_
                    else:
                        obj = value

            else:
                obj = _not_found_
        else:
            obj = Path._apply_op(_op, obj, *args, **kwargs)

        if obj is _not_found_:
            obj = default_value

        # if idx < len(path) - 1 and obj is _not_found_:
        #     raise KeyError(f"Path not found {path[:idx + 1]}")

        return obj

    @staticmethod
    def _do_for_each(source, path, *args, **kwargs) -> typing.Generator[typing.Any, None, None]:
        """遍历容器子节点。
        @NOTE:
            - 叶节点无输出
            - Sequence: 输出 value 子节点
            - Mapping:  输出 key和 value
        @FIXME:
            - level 参数，用于实现多层遍历，尚未实现
        """

        if path is None:
            suffix = None
        else:
            for idx, p in enumerate(path):
                if p is Path.tags.children:
                    source = Path._do_find(source, path[:idx])
                    suffix = path[idx + 1 :]
                    break
                elif isinstance(p, Path.tags):
                    raise NotImplementedError(p)
            else:
                source = Path._do_find(source, path)
                suffix = []

        if isinstance(source, collections.abc.Sequence) and not isinstance(source, str):
            for k, v in enumerate(source):
                yield k, Path._do_find(v, suffix, *args, **kwargs)

        elif isinstance(source, collections.abc.Mapping):
            for k, v in source.items():
                yield k, Path._do_find(v, suffix, *args, **kwargs)

        elif source is not _not_found_:
            logger.warning(f"{type(source)} is not iterable!")

    ####################################################
    # operation

    @staticmethod
    def _op_is_leaf(source: typing.Any, *args, **kwargs) -> bool:
        return not isinstance(source, (collections.abc.Mapping, collections.abc.Sequence))

    @staticmethod
    def _op_is_list(source: typing.Any, *args, **kwargs) -> bool:
        return isinstance(source, collections.abc.Sequence)

    @staticmethod
    def _op_is_dict(source: typing.Any, *args, **kwargs) -> bool:
        return isinstance(source, collections.abc.Mapping)

    @staticmethod
    def _op_check_type(source: typing.Any, tp, *args, **kwargs) -> bool:
        return isinstance_generic(source, tp)

    @staticmethod
    def _op_count(source: typing.Any, *args, **kwargs) -> int:
        if source is _not_found_:
            return 0
        elif (
            isinstance(source, collections.abc.Sequence) or isinstance(source, collections.abc.Mapping)
        ) and not isinstance(source, str):
            return len(source)
        else:
            return 1

    @staticmethod
    def _op_fetch(source: typing.Any, *args, default_value=_not_found_, **kwargs) -> bool:
        if source is _not_found_:
            source = default_value
        return source

    @staticmethod
    def _op_exists(source: typing.Any, *args, **kwargs) -> bool:
        return source is not _not_found_

    @staticmethod
    def _op_search(source: typing.Iterable, *args, search_range=None, **kwargs):
        if not isinstance(source, collections.abc.Sequence):
            raise TypeError(f"{type(source)} is not sequence")

        query = Query(*args, **kwargs)

        if search_range is not None and isinstance(source, collections.abc.Sequence):
            source = source[search_range]

        for idx, value in enumerate(source):
            if query.check(value):
                break
        else:
            idx = None
            value = _not_found_

        return idx, value

    @staticmethod
    def _op_call(source, *args, **kwargs) -> typing.Any:
        if not callable(source):
            logger.error(f"Not callable! {source}")
            return _not_found_
        else:
            return source(*args, **kwargs)

    @staticmethod
    def _op_next(target, query, start: int | None = None, *args, **kwargs) -> typing.Tuple[typing.Any, int | None]:
        if not isinstance(query, (slice, set, Query)):
            raise ValueError(f"query is not dict,slice! {query}")

        if target is _not_found_ or target is None:
            return _not_found_, None

        if isinstance(query, slice):
            if start is None or start is _not_found_:
                start = query.start or 0
            elif query.start is not None and start < query.start:
                raise IndexError(f"Out of range: {start} < {query.start}!")
            stop = query.stop or len(target)
            step = query.step or 1

            if start >= stop:
                # raise StopIteration(f"Can not find next entry of {start}>={stop}!")
                return None, None
            else:
                value = Path._op_fetch(target, start, *args, default_value=_not_found_, **kwargs)

                if value is _not_found_:
                    start = None
                else:
                    start += step

                return value, start

        elif isinstance(query, Query):
            if start is None or start is _not_found_:
                start = 0

            stop = len(target)

            value = _not_found_

            while start < stop:
                value = target[start]
                if not Path._op_check(value, query, *args, **kwargs):
                    start += 1
                    continue
                else:
                    break

            if start >= stop:
                return _not_found_, None
            else:
                return value, start

        else:
            raise NotImplementedError(f"Not implemented yet! {type(query)}")


PathLike = str | int | slice | dict | list | OpTags | Path | None

path_like = [int, str, slice, list, None, tuple, set, dict, OpTags, Path]


_T = typing.TypeVar("_T")


def update_tree(target: _T, *args, **kwargs) -> _T:
    for d in [*args, kwargs]:
        target = Path._do_update(target, [], d)
    return target


def merge_tree(*args, **kwargs) -> _T:
    return update_tree({}, *args, **kwargs)


def as_path(path):
    if path is None or path is _not_found_:
        return Path()
    elif not isinstance(path, Path):
        return Path(path)
    else:
        return path



import collections.abc
import typing

import numpy as np

from ..utils.typing import ArrayType
from .GeoObject import GeoObject
from .PointSet import PointSet


@GeoObject.register(["polyline", "Polyline"])
class Polyline(PointSet):

    def __init__(self, *args,  **kwargs) -> None:
        super().__init__(*args, rank=1, **kwargs)



from __future__ import annotations

import collections.abc
import typing
import numpy as np
from functools import cached_property
from enum import Enum

from ..geometry.GeoObject import GeoObject, GeoObjectSet, as_geo_object
from ..core.Domain import DomainBase
from ..core.Path import update_tree

from ..utils.logger import logger
from ..utils.misc import group_dict_by_prefix
from ..utils.plugin import Pluggable
from ..utils.typing import ArrayType, NumericType, ScalarType, as_array
from ..utils.tags import _not_found_


class Mesh(DomainBase, Pluggable):
    """Mesh  网格

    @NOTE: In general, a mesh provides more flexibility in representing complex geometries and
    can adapt to the local features of the solution, while a grid is simpler to generate
    and can be more efficient for certain types of problems.
    """

    _plugin_registry = {}
    _plugin_prefix = "spdm.mesh.mesh_"

    def __init__(self, *args, **kwargs) -> None:
        if self.__class__ is Mesh:
            if len(args) == 1:
                if isinstance(args[0], (list, tuple)):
                    args = tuple(args[0])
                elif isinstance(args[0], dict):
                    kwargs = update_tree(args[0], kwargs)
                    args = tuple([])

            mesh_type = kwargs.pop("type", None)

            if mesh_type is not _not_found_ or mesh_type is None or not mesh_type:
                pass
            elif isinstance(mesh_type, Enum):
                mesh_type = mesh_type.name

            if len(args) == 0:
                dims, kwargs = group_dict_by_prefix(kwargs, prefixes="dim", sep=None)
                if dims is None:
                    pass
                elif len(dims) == 1 and "s" in dims:
                    dims = dims["s"]
                else:
                    dims = {int(k): v for k, v in dims.items() if k.isdigit()}
                    dims = dict(sorted(dims.items(), key=lambda x: x[0]))
                    dims = tuple([as_array(d) for d in dims.values()])
            else:
                dims = args

            if mesh_type is None or mesh_type is _not_found_ and dims is not None:
                ndim = len(dims)

                if all([isinstance(d, np.ndarray) for d in dims]):
                    if all([d.ndim == 1 for d in dims]):
                        mesh_type = "rectilinear"
                    elif all([d.ndim == ndim for d in dims]):
                        mesh_type = "rectangular"
                    else:
                        raise RuntimeError(f"illegal dims {dims}")

            if mesh_type is not None:
                super().__dispatch_init__(mesh_type, self, *dims, **kwargs)
            else:
                logger.warning(f"Can not determint mesh type!")

            return

        geometry, kwargs = group_dict_by_prefix(kwargs, "geometry")

        if isinstance(geometry, Enum):
            geometry = {"type": geometry.name}
        elif isinstance(geometry, str):
            geometry = {"type": geometry}

        if isinstance(geometry, (GeoObject, GeoObjectSet)):
            self._geometry = geometry
        elif isinstance(geometry, collections.abc.Mapping):
            self._geometry = GeoObject(*args, **geometry)
        else:
            raise RuntimeError(f"Mesh.__init__(): geometry={geometry} is not found")

        self._shape: ArrayType = np.asarray(self._metadata.get("shape", []), dtype=int)

        DomainBase.__init__(self, *args, **kwargs)

    @property
    def axis_label(self) -> typing.Tuple[str]:
        return self._metadata.get("axis_label", ["[-]"] * self.ndim)

    @property
    def name(self) -> str:
        return self._metadata.get("name", "unamed")

    @property
    def type(self) -> str:
        return self._metadata.get("type", "unknown")

    @property
    def units(self) -> typing.Tuple[str, ...]:
        return tuple(self._metadata.get("units", ["-"]))

    @property
    def geometry(self) -> GeoObject:
        """Geometry of the Mesh  网格的几何形状"""
        return self._geometry

    @property
    def ndim(self) -> int:
        return self.geometry.ndim

    @property
    def rank(self) -> int:
        return self.geometry.rank

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """
        存储网格点数组的形状
        TODO: support multiblock Mesh
        结构化网格 shape   如 [n,m] n,m 为网格的长度dimension
        非结构化网格 shape 如 [<number of vertices>]
        """
        return self._shape

    def parametric_coordinates(self, *xyz) -> ArrayType:
        """parametric coordinates

        网格点的 _参数坐标_
        Parametric coordinates, also known as computational coordinates or intrinsic coordinates,
        are a way to represent the position of a point within an element of a mesh.
        一般记作 u,v,w \in [0,1] ,其中 0 表示“起点”或 “原点” origin，1 表示终点end
        mesh的参数坐标(u,v,w)，(...,0)和(...,1)表示边界

        @return: 数组形状为 [geometry.rank, <shape of xyz ...>]的数组
        """
        if len(xyz) == 0:
            return np.stack(np.meshgrid(*[np.linspace(0.0, 1.0, n, endpoint=True) for n in self.shape]))
        else:
            raise NotImplementedError(f"{self.__class__.__name__}.parametric_coordinates for unstructured mesh")

    def coordinates(self, *uvw) -> ArrayType:
        """网格点的 _空间坐标_
        @return: _数组_ 形状为 [<shape of uvw ...>,geometry.ndim]
        """
        return self.geometry.coordinates(uvw if len(uvw) > 0 else self.parametric_coordinates())

    def uvw(self) -> ArrayType:
        return self.parametric_coordinates(*xyz)

    """ alias of parametric_coordiantes"""

    @cached_property
    def vertices(self) -> ArrayType:
        """coordinates of vertice of mesh  [<shape...>, geometry.ndim]"""
        return self.geometry.coordinates(self.parametric_coordinates())

    @cached_property
    def points(self) -> typing.List[ArrayType]:
        """alias of vertices, change the shape to tuple"""
        return [self.vertices[..., idx] for idx in range(self.ndim)]

    @cached_property
    def xyz(self) -> typing.List[ArrayType]:
        return self.points

    @property
    def cells(self) -> typing.Any:
        raise NotImplementedError(f"{self.__class__.__name__}.cells")

    """ refer to the individual units that make up the mesh"""

    def interpolator(self, y: NumericType, *args, **kwargs) -> typing.Callable[..., NumericType]:
        raise NotImplementedError(f"{self.__class__.__name__}.interpolator")

    def partial_derivative(self, order, y: NumericType, *args, **kwargs) -> typing.Callable[..., NumericType]:
        raise NotImplementedError(f"{self.__class__.__name__}.partial_derivative")

    def antiderivative(self, y: NumericType, *args, **kwargs) -> typing.Callable[..., NumericType]:
        raise NotImplementedError(f"{self.__class__.__name__}.antiderivative")

    def integrate(self, y: NumericType, *args, **kwargs) -> ScalarType:
        raise NotImplementedError(f"{self.__class__.__name__}.integrate")

    def eval(self, func, *args, **kwargs) -> ArrayType:
        return func(*self.points)


@Mesh.register(["null", None])
class NullMesh(Mesh):
    def __init__(self, *args, **kwargs) -> None:
        if len(args) > 0 or len(kwargs) > 0:
            raise RuntimeError(f"Ignore args {args} and kwargs {kwargs}")
        super().__init__()


@Mesh.register("regular")
class RegularMesh(Mesh):
    pass


def as_mesh(*args, **kwargs) -> Mesh:
    if len(args) == 1 and isinstance(args[0], Mesh):
        if len(kwargs) > 0:
            logger.warning(f"Ignore kwargs {kwargs}")
        return args[0]
    else:
        return Mesh(*args, **kwargs)



import abc
import collections.abc
import typing
from functools import cached_property
from typing import Any

import numpy as np

from ..utils.logger import logger
from ..utils.typing import ArrayType
from .BBox import BBox
from .GeoObject import GeoObject
from .Point import Point
from .Vector import Vector


@GeoObject.register("line")
class Line(GeoObject):
    """ Line
        线，一维几何体
    """

    def __init__(self, p0: Point | ArrayType, p1: Point | ArrayType, *args, **kwargs) -> None:
        super().__init__(*args, rank=1,   **kwargs)
        self._p0 = Point(*p0)
        self._p1 = Point(*p1)

    @property
    def p0(self) -> Point: return self._p0

    @property
    def p1(self) -> Point: return self._p1

    @property
    def length(self) -> float: return self.measure

    @property
    def measure(self) -> float: return np.Inf

    @property
    def direction(self) -> Vector: return Vector(self.p1-self.p0)

    @property
    def boundary(self) -> typing.List[Point]: return [self.p0, self.p1]

    @property
    def bbox(self) -> BBox: return BBox([self.p0.x,self.p0.y], [self.p1.x-self.p0.x, self.p1.y-self.p0.y])
    """ boundary box of geometry [ [...min], [...max] ] """

    def contains(self, o) -> bool:
        return self._impl.contains(o._impl if isinstance(o, GeoObject) else o)

    def coordinates(self, u: float | np.ndarray | typing.List[np.ndarray] | None = None) -> ArrayType:
        direction = self.direction
        if u is None:
            return [self.p0, self.p1]
        elif isinstance(u, float):
            return (self.p0 + self.direction*u)
        elif isinstance(u, np.ndarray):
            return np.asarray([(self.p0 + self.direction*v)[:] for v in u])
        elif isinstance(u, collections.abc.Iterable):
            return np.asarray([(self.p0 + direction*v) for v in u])
        else:
            raise RuntimeError(f"Invalid input type {type(u)}")

    def project(self, o) -> Point:
        return Point(self._impl.projection(o._impl if isinstance(o, GeoObject) else o))

    def bisectors(self, o) -> typing.List[GeoObject]:
        return [GeoObject(v) for v in self._impl.bisectors(o._impl if isinstance(o, GeoObject) else o)]

    @cached_property
    def dl(self) -> np.ndarray:
        x, y = np.moveaxis(self.points, -1, 0)

        a, b = self.derivative()

        # a = a[:-1]
        # b = b[:-1]
        dx = x[1:]-x[:-1]
        dy = y[1:]-y[:-1]

        m1 = (-a[:-1]*dy+b[:-1]*dx)/(a[:-1]*dx+b[:-1]*dy)

        # a = np.roll(a, 1, axis=0)
        # b = np.roll(b, 1, axis=0)

        m2 = (-a[1:]*dy+b[1:]*dx)/(a[1:]*dx+b[1:]*dy)

        return np.sqrt(dx**2+dy**2)*(1 + (2.0*m1**2+2.0*m2**2-m1*m2)/30)

    def integral(self, func: typing.Callable) -> float:
        x, y = self.xyz
        val = func(x, y)
        # c_pts = self.points((self._mesh[0][1:] + self._mesh[0][:-1])*0.5)

        return np.sum(0.5*(val[:-1]+val[1:]) * self.dl)

    # def average(self, func: Callable[[_TCoord, _TCoord], _TCoord]) -> float:
    #     return self.integral(func)/self.length

    def encloses_point(self, *x: float, **kwargs) -> bool:
        return super().enclosed(**x, **kwargs)

    def trim(self):
        return NotImplemented

    def remesh(self, mesh_type=None):
        return NotImplemented


@Line.register("ray")
class Ray(Line):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


@Line.register("segment")
class Segment(Line):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def midpoint(self) -> Point:
        return Point((self.p0+self.p1)*0.5)



import collections
import os
import typing

import MDSplus as mds
import numpy as np
from spdm.core.Collection import Collection
from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.utils.logger import logger
from spdm.utils.uri_utils import URITuple
from spdm.utils.tags import _not_found_


@File.register(["mdsplus", "mds", "mds+"])
class MDSplusTree(File):
    MDS_MODE = {
        File.Mode.read: "ReadOnly",
        File.Mode.write: "Normal",
        File.Mode.read | File.Mode.write: "Normal",
        File.Mode.read | File.Mode.write | File.Mode.create: "Edit",
        File.Mode.write | File.Mode.create: "New",
    }

    def __init__(self, *args, fid=None, shot: typing.Optional[int] = None, tree_name: typing.Optional[str] = None, **kwargs):
        super().__init__(*args, **kwargs)

        self._envs = {}
        # {k: (v if not isinstance(v, slice) else f"{v.start}:{v.stop}:{v.step}")
        #   for k, v in self._envs.items()}

        query = self.url.query or {}

        self._mds_mode = MDSplusTree.MDS_MODE[self.mode]

        path = self.url.path.rstrip("/")

        if self.url.authority != "":
            path = f"{self.url.authority}::{path}"

        if self.url.protocol in ("ssh"):
            path = f"{self.url.protocol}://" + path

        if len(path) == 0:
            path = None

        self._default_tree_path = path

        if tree_name is None:
            tree_name = query.get("tree_name", None)

        self._old_env = {}

        if tree_name is None:
            self._default_tree_name = None
        elif self._default_tree_path is not None:
            tree_name = tree_name.split(",")
            self._default_tree_name = tree_name[0]
            for p in tree_name:
                env_path = f"{p}_path"
                if env_path in os.environ:
                    self._old_env[env_path] = os.environ[env_path]
                    os.environ[f"{p}_path"] = f"{path}:{self._old_env[env_path]}"
                else:
                    self._old_env[env_path] = None
                    os.environ[f"{p}_path"] = path

        self._shot = (shot if shot is not None
                      else (fid if fid is not None else query.get("shot", self.url.fragment)))

        self._trees = {}

        self._entry = MDSplusEntry(self)

    def __del__(self):
        # del self._trees
        for k, tree in self._trees.items():
            tree.close()
            logger.debug(f"Close MDS Tree:{k}")

        self._trees = {}

        for k, v in self._old_env.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v

    @property
    def entry(self, lazy=True) -> Entry:
        return self._entry

    def get_tree(self,
                 tree_name: typing.Optional[str] = None,
                 tree_path: typing.Optional[str] = None,
                 ):
        if tree_name is None:
            tree_name = self._default_tree_name

        if tree_name in self._trees:
            return self._trees[tree_name]

        mode = self._mds_mode

        shot = int(self._shot) if isinstance(self._shot, str) else self._shot

        if tree_path is None:
            tree_path = self._default_tree_path

        try:
            tree = mds.Tree(tree_name, shot, mode=mode, path=tree_path)
        except mds.mdsExceptions.TreeFOPENR as error:
            raise FileNotFoundError(
                f"Can not open mdsplus tree! tree_name={tree_name} shot={shot} tree_path={tree_path} mode={mode} \n {error}")
        except mds.mdsExceptions.TreeNOPATH as error:
            raise FileNotFoundError(f"{tree_name}_path is not defined! tree_name={tree_name} shot={shot}  \n {error}")
        else:
            logger.debug(f"Open MDSplus Tree [{tree_name}] shot={shot}")

        self._trees[tree_name] = tree

        return tree

    def query(self, request, prefix=None, **kwargs) -> typing.Any:
        if request is None:
            return _not_found_

        if isinstance(request, str):
            request = {"query": request}

        request = collections.ChainMap(request, kwargs)

        # elif isinstance(request, collections.abc.Mapping):

        tree_name = request.get("@tree", self._default_tree_name)
        tree_path = request.get("@tree_path", None)

        tdi = request.get("query", None) or request.get("@text", None)

        if not tdi:
            raise RuntimeError(f"Illegal request {request}")

        try:
            tdi = tdi.format_map(self._envs)
        except KeyError as error:
            raise KeyError(f"Can not format tdi! {error} tdi={tdi} envs={self._envs} prefix={prefix}")

        res = None
        tree = self.get_tree(tree_name, tree_path)
        try:
            res = tree.tdiExecute(tdi).data()
        except mds.mdsExceptions.TdiException as error:
            # raise RuntimeError(f"MDSplus TDI error [{tdi}]! {error}")
            logger.warning(f'MDS TDI error! tree_name={tree_name} shot={self._shot} tdi="{tdi}" \n {error}')

        except mds.mdsExceptions.TreeNODATA as error:
            logger.warning(f'MDS No data! tree_name={tree_name} shot={self._shot} tdi="{tdi}" \n {error}')

        except Exception as error:
            raise RuntimeError(f'mds.mdsExceptions! tree_name={tree_name} shot={self._shot} tdi="{tdi}"') from error
            # raise error

        if not isinstance(res, np.ndarray):
            pass
        elif len(res.shape) == 2:
            if res.shape[1] == 1:
                res = res[:, 0]
            elif res.shape[0] == 1:
                res = res[0]
            else:
                res = res.transpose(1, 0)
        return res

    def update(self, *args, envs=None, **kwargs):
        raise NotImplementedError()


@Collection.register(["mdsplus", "mds", "mds+", "MDSplus"])
class MDSplusCollection(Collection):
    def insert_one(self, fid=None, *args, query=None, mode=None, **kwargs):
        fid = (
            fid
            or self.guess_id(*args, **collections.ChainMap((query or {}), kwargs))
            or self.next_id
        )
        return MDSplusTree(self.url, fid=fid, mode=mode or "w", **kwargs)

    def search_one(self, predicate, projection=None, only_one=False, **kwargs) -> Entry:
        fid = self.guess_id(predicate, **kwargs)
        entry = self._mapping(MDSplusTree(self.url, fid=fid, **kwargs).entry)
        if projection is None:
            return entry
        else:
            return entry.find(projection)

    def count(self, predicate=None, *args, **kwargs) -> int:
        return NotImplemented()

    # def search_one(self, predicate: Document,  projection: Document = None, *args, **kwargs):
    #     shot = getitem(predicate, "shot", None) or getitem(predicate, "_id", None)
    #     if shot is not None:
    #         return MDSplusEntry(self._tree_name, shot, mode="r") .find(projection)
    #     else:
    #         for shot in self._foreach_shot():
    #             res = MDSplusEntry(self._tree_name, shot, mode="r").find_if(
    #                 projection, predicate)
    #             if res is not None:
    #                 return res
    #     return None

    # def _foreach_shot(self):
    #     f_prefix = f"{self._tree_name.lower()}_"
    #     f_prefix_l = len(f_prefix)
    #     glob = f"{f_prefix}*.tree"
    #     for fp in self._path.glob(glob):
    #         yield fp.stem[f_prefix_l:]

    # def search(self, predicate: Document = None, projection: Document = None, *args, **kwargs):

    #     for shot in self._foreach_shot():
    #         res = MDSplusEntry(self._tree_name, shot, mode="r").find_if(projection, predicate)
    #         logger.debug(res)

    #         if res is not None:
    #             yield res

    # def insert_one(self, document: Document, *args, **kwargs):
    #     self._count += 1

    #     shot = int(document.get("shot", self._count))

    #     MDSplusEntry(self._tree_name, shot, mode="x").update(document)

    #     return shot


@Entry.register(["mdsplus", "mds"])
class MDSplusEntry(Entry):
    def __init__(self, cache:  MDSplusTree | str | URITuple, *args, **kwargs):
        if isinstance(cache, (str, URITuple)):
            cache: MDSplusTree = MDSplusTree(cache)
        if not isinstance(cache, MDSplusTree):
            raise TypeError(f"cache must be MDSplusEntry or str, but got {type(cache)}")
        super().__init__(cache, *args, **kwargs)

    def find(self, request, *args, **kwargs):
        return self._data.query(request, *args, prefix=self._path, **kwargs)

    def update(self, *args, **kwargs):
        return self._data.update(*args, prefix=self._path, **kwargs)

    def search(self, *args, **kwargs):
        yield from self._data.search(*args, prefix=self._path, **kwargs)


def open_mdstree(tree_name, shot, mode="NORMAL", path=None):
    if tree_name is None:
        raise ValueError(f"Treename is empty!")
    try:
        shot = int(shot)
        logger.info(
            f"Open MDSTree: tree_name={tree_name} shot={shot} mode=\"{mode}\" path='{path}'"
        )
        tree = mds.Tree(tree_name, shot, mode=mode, path=path)
    except mds.mdsExceptions.TreeFOPENR as error:
        # tree_path = os.environ.get(f"{tree_name}_path", None)
        raise FileNotFoundError(
            f"Can not open mdsplus tree! tree_name={tree_name} shot={shot} tree_path={path} mode={mode} \n {error}"
        )
    except mds.mdsExceptions.TreeNOPATH as error:
        raise FileNotFoundError(
            f"{tree_name}_path is not defined! tree_name={tree_name} shot={shot}  \n {error}"
        )
    return tree


__SP_EXPORT__ = MDSplusTree



from spdm.data.Expression import Expression, Variable
from spdm.data.Functor import Functor
from spdm.numlib.calculus import derivative
from spdm.data.Function import Function
import numpy as np
from spdm.utils.logger import logger


if __name__ == '__main__':
    _x = Variable(0, "x")
    _y = Variable(1, "y")
    expr = derivative(1, (np.cos(_x)+np.sin(_y))*2)
    logger.debug(expr._repr_latex_())



import typing

import numpy as np

from ..utils.typing import ArrayLike, ArrayType, as_array
from .Mesh import Mesh


class StructuredMesh(Mesh):
    """ StructureMesh
      
    结构化网格上的点可以表示为长度为n=rank的归一化ntuple，记作 uv，uv_r \in [0,1]

    """

    def __init__(self, shape: ArrayLike, *args, cycles=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._shape = as_array(shape)
        self._cycles = cycles
        # shape = tuple(shape)
        # if cycles is None:
        #     cycles = [False] * ndims
        # if not isinstance(cycles, collections.abc.Sequence):
        #     cycles = [cycles]
        # if len(cycles) == 1:
        #     cycles = cycles * ndims

        # self._cycles: typing.Tuple[int] = self._metadata.get("cycles", None)

        # if self._cycles is None:
        #     self._cycles = ([False]*self.geometry.rank)
        # # logger.debug(f"Create {self.__class__.__name__} rank={self.rank} shape={self.shape} ndims={self.ndims}")
        # bbox = self.geometry.bbox
        # shape = self.shape
        # if not isinstance(shape, np.ndarray):
        #     raise TypeError(f"shape is not np.ndarray")
        # self._dx = (bbox[1]-bbox[0])/shape
        # self._origin = bbox[0]

    @property
    def cycles(self) -> typing.List[float]: return self._cycles
    """ Periodic boundary condition   周期性边界条件,  标识每个维度是否是周期性边界 """

    @property
    def origin(self) -> ArrayType: return self._origin

    @property
    def dx(self) -> ArrayType: return self._dx

    def coordinates(self, *uvw) -> ArrayType:
        if len(uvw) == 1:
            uvw = uvw[0]
        return np.stack([((uvw[i])*self.dx[i]+self.origin[i]) for i in range(self.rank)])

    def parametric_coordinates(self, *xyz) -> ArrayType:
        if len(uvw) == 1:
            uvw = uvw[0]
        return np.stack([((xyz[i]-self.origin[i])/self.dx[i]) for i in range(self.rank)])

    def interpolator(self, *args, **kwargs) -> typing.Callable:
        """ Interpolator of the Mesh
            网格的插值器, 用于网格上的插值
            返回一个函数，该函数的输入是一个坐标，输出是一个值
            输入坐标若为标量，则返回标量值
            输入坐标若为数组，则返回数组
        """
        raise NotImplementedError(f"{args} {kwargs}")



import unittest

import numpy as np
from scipy import constants
from spdm.core.Expression import Expression, Variable, derivative, antiderivative
from spdm.core.Function import Function
from spdm.utils.logger import logger

TWOPI = constants.pi * 2.0


class TestFunction(unittest.TestCase):
    def test_expression(self):
        x = np.linspace(0, 1.0, 128)
        y = np.sin(x * TWOPI)

        fun = Function(x, y)
        expr = fun * 2.0
        self.assertTrue(type(expr) is Expression)

    def test_operator(self):
        x = np.linspace(0, 1, 128)
        y = np.linspace(0, 2, 128)
        fun = Function(x, y)

        self.assertTrue(np.all(-fun == -y))
        self.assertTrue(np.all(fun + 2 == y + 2))
        self.assertTrue(np.all(fun - 2 == y - 2))
        self.assertTrue(np.all(fun * 2 == y * 2))
        self.assertTrue(np.all(fun / 2 == y / 2))
        self.assertTrue(np.all(fun**2 == y**2))
        # self.assertTrue(np.all(fun @ fun == y)

    def test_construct_from_expression(self):
        x = np.linspace(0, 1, 128)
        y = np.linspace(0, 2, 128)
        fun = Function(x, y * 2)

        self.assertTrue(np.allclose(fun, y * 2))

    def test_np_fun(self):
        x = np.linspace(0, 1, 128)
        y = np.linspace(0, 2, 128)

        fun = Function(x, y)

        self.assertTrue(type(fun + 1) is Expression)
        self.assertTrue(type(fun * 2) is Expression)
        self.assertTrue(type(np.sin(fun)) is Expression)

    def test_spl(self):
        x = np.linspace(0, 1.0, 128)
        y = np.sin(x * TWOPI)

        fun = Function(x, y)

        x2 = np.linspace(0, 1.0, 64)

        y2 = np.sin(x2 * TWOPI)

        self.assertTrue(np.allclose(y2, fun(x2)))

    def test_antiderivative(self):
        _x = Variable(0, "x")
        sinx = derivative(np.sin(_x), _x)
        cosx = np.cos(_x)

        x = np.linspace(0, TWOPI, 128)
        self.assertTrue(np.allclose(sinx(x), cosx(x)))


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import collections.abc
import typing
import uuid
from copy import copy
from functools import cached_property

import numpy as np

from ..core.HTree import List
from ..utils.logger import logger
from ..utils.plugin import Pluggable
from ..utils.typing import ArrayLike, ArrayType, NumericType, ScalarType, array_type, nTupleType, numeric_type
from .BBox import BBox


class GeoObject(Pluggable):
    """Geomertic object
    几何对象，包括点、线、面、体等

    TODO:
        - 支持3D可视化 （Jupyter+？）

    """

    _plugin_module_path_template = "spdm.geometry.{name}"
    _plugin_registry = {}

    @classmethod
    def __dispatch_init__(cls, _geo_type, self, *args, **kwargs) -> None:
        """ """
        if _geo_type is None or len(_geo_type) == 0:
            _geo_type = kwargs.pop("type", None)

        # if isinstance(_geo_type, str):
        #     _geo_type = [_geo_type,
        #                  f"spdm.geometry.{_geo_type}#{_geo_type}",
        #                  f"spdm.geometry.{_geo_type}{cls.__name__}#{_geo_type}{cls.__name__}",
        #                  f"spdm.geometry.{_geo_type.capitalize()}#{_geo_type.capitalize()}",
        #                  f"spdm.geometry.{_geo_type.capitalize()}{cls.__name__}#{_geo_type.capitalize()}{cls.__name__}",
        #                  f"spdm.geometry.{cls.__name__}#{_geo_type}"
        #                  ]

        super().__dispatch_init__(_geo_type, self, *args, **kwargs)

    def __init__(self, *args, ndim: int = 0, rank: int = -1, **kwargs) -> None:
        if self.__class__ is GeoObject:
            GeoObject.__dispatch_init__(None, self, *args, ndim=ndim, rank=rank, **kwargs)
            return

        self._metadata = kwargs.pop("metadata", {})
        self._metadata.update(kwargs)
        self._metadata.setdefault("name", f"{self.__class__.__name__}_{uuid.uuid1()}")

        self._ndim = ndim
        self._rank = rank if rank >= 0 else ndim

    def __copy__(self) -> GeoObject:
        other: GeoObject = object.__new__(self.__class__)
        other._metadata = copy(self._metadata)
        other._ndim = self._ndim
        other._rank = self._rank
        return other
        # return self.__class__(rank=self.rank, ndim=self.ndim, **self._metadata)

    # def _repr_html_(self) -> str:
    #     """Jupyter 通过调用 _repr_html_ 显示对象"""
    #     from ..view.View import display
    #     return display(self, schema="html")

    def _repr_svg_(self) -> str:
        """Jupyter 通过调用 _repr_html_ 显示对象"""

        from ..view.View import display

        return display(self, schema="svg")

    def __equal__(self, other: GeoObject) -> bool:
        return (
            isinstance(other, GeoObject)
            and self.rank == other.rank
            and self.ndim == other.ndim
            and self.bbox == other.bbox
        )

    @property
    def name(self) -> str:
        return self._metadata.get("name", "unnamed")

    @property
    def rank(self) -> int:
        return self._rank

    """ 几何体（流形）维度  rank <=ndims

            0: point
            1: curve
            2: surface
            3: volume
            >=4: not defined
        The rank of a geometric object refers to the number of independent directions
        in which it extends. For example, a point has rank 0, a line has rank 1,
        a plane has rank 2, and a volume has rank 3.
    """

    @property
    def number_of_dimensions(self) -> int:
        return self._ndim

    """ 几何体所处的空间维度， = 0，1，2，3 ,...
        The dimension of a geometric object, on the other hand, refers to the minimum number of
        coordinates needed to specify any point within it. In general, the rank and dimension of
        a geometric object are the same. However, there are some cases where they can differ.
        For example, a curve that is embedded in three-dimensional space has rank 1 because
        it extends in only one independent direction, but it has dimension 3 because three
        coordinates are needed to specify any point on the curve.
    """

    @property
    def ndim(self) -> int:
        return self._ndim

    """ alias of dimension """

    @property
    def boundary(self) -> GeoObject | None:
        """boundary of geometry which is a geometry of rank-1"""
        if self.is_closed:
            return None
        else:
            raise NotImplementedError(f"{self.__class__.__name__}.boundary")

    @property
    def is_convex(self) -> bool:
        return self._metadata.get("convex", True)

    """ is convex """

    @property
    def is_closed(self) -> bool:
        return self._metadata.get("closed", True)

    @property
    def bbox(self) -> BBox:
        raise NotImplementedError(f"{self.__class__.__name__}.bbox")

    """ boundary box of geometry [ [...min], [...max] ] """

    @property
    def measure(self) -> float:
        return self.bbox.measure

    """ measure of geometry, length,area,volume,etc. 默认为 bbox 的体积 """

    def enclose(self, *args) -> bool | array_type:
        """Return True if all args are inside the geometry, False otherwise."""
        return False if not self.is_closed else self.bbox.enclose(*args)

    def intersection(self, other: GeoObject) -> typing.List[GeoObject]:
        """Return the intersection of self with other."""
        return [self.bbox.intersection(other.bbox)]

    def reflect(self, point0, point1) -> GeoObject:
        """reflect  by line"""
        other = copy(self)
        other._metadata["name"] = f"{self.name}_reflect"
        other.bbox.reflect(point0, point1)
        return other

    def rotate(self, angle, axis=None) -> GeoObject:
        """rotate  by angle and axis"""
        other = copy(self)
        other._metadata["name"] = f"{self.name}_rotate"
        other.bbox.rotate(angle, axis=axis)
        return other

    def scale(self, *s, point=None) -> GeoObject:
        """scale self by *s, point"""
        other = copy(self)
        other._metadata["name"] = f"{self.name}_scale"
        other.bbox.scale(*s, point=point)
        return other

    def translate(self, *shift) -> GeoObject:
        other = copy(self)
        other._metadata["name"] = f"{self.name}_translate"
        other.bbox.translate(*shift)
        return other

    def trim(self):
        raise NotImplementedError(f"{self.__class__.__name__}.trim")

    @staticmethod
    def _normal_points(*args) -> np.ndarray | typing.List[float]:
        if len(args) == 0:
            return []
        elif len(args) == 1:
            return args[0]
        elif isinstance(args[0], (int, float, bool, complex)):
            return list(args)
        elif isinstance(args[0], collections.abc.Sequence):
            return np.asarray([GeoObject._normal_points(*p) for p in args])
        else:
            raise TypeError(f"args has wrong type {type(args[0])} {args}")


_TG = typing.TypeVar("_TG")


class GeoObjectSet(typing.List[GeoObject]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args)
        rank = kwargs.pop("rank", None)
        ndim = kwargs.pop("ndim", None)
        if rank is None:
            ranks = [obj.rank for obj in self if isinstance(obj, GeoObject)]
            rank = max(ranks) if len(ranks) > 0 else 0

        if ndim is None:
            ndim_list = [obj.ndim for obj in self if isinstance(obj, GeoObject)]
            if len(ndim_list) > 0:
                ndim = max(ndim_list)  # [0]
            else:
                raise RuntimeError(f"Can not get ndim from {ndim_list}")
        self._rank = rank
        self._ndim = ndim
        self._metadata = kwargs
        # GeoObject.__init__(self, rank=rank, ndim=ndim, **kwargs)

    def _repr_svg_(self) -> str:
        """Jupyter 通过调用 _repr_html_ 显示对象"""
        from ..view.View import display

        return display(self, schema="svg")

    # def __svg__(self) -> str:
    #     return f"<g >\n" + "\t\n".join([g.__svg__ for g in self if isinstance(g, GeoObject)]) + "</g>"

    @property
    def bbox(self) -> BBox:
        return np.bitwise_or.reduce([g.bbox for g in self if isinstance(g, GeoObject)])

    def enclose(self, other) -> bool:
        return all([g.enclose(other) for g in self if isinstance(g, GeoObject)])

    # class Box(GeoObject):
    #     def __init__(self, *args, **kwargs) -> None:
    #         super().__init__(*args, **kwargs)

    #     @property
    #     def bbox(self) -> typing.Tuple[ArrayType, ArrayType]: return self._points[0], self._points[1]

    #     def enclose(self, *xargs) -> bool:
    #         if all([isinstance(x, numeric_type) for x in xargs]):  # 点坐标
    #             if len(xargs) != self.ndim:
    #                 raise RuntimeError(f"len(xargs)={len(xargs)}!=self.ndim={self.ndim} {xargs}")
    #             xmin, xmax = self.bbox
    #             return np.bitwise_and.reduce([((xargs[i] >= xmin[i]) & (xargs[i] <= xmax[i])) for i in range(self.ndim)])
    #         elif len(xargs) == 1 and isinstance(xargs[0], GeoObject):
    #             raise NotImplementedError(f"{self.__class__.__name__}.enclose(GeoObject)")
    #         else:
    #             return np.bitwise_and.reduce([self.enclose(x) for x in xargs])


def as_geo_object(*args, **kwargs) -> GeoObject | GeoObjectSet:
    if len(kwargs) > 0 or len(args) != 1:
        return GeoObject(*args, **kwargs)
    elif isinstance(args[0], GeoObject):
        return args[0]
    elif isinstance(args[0], collections.abc.Sequence):
        return GeoObjectSet(*args, **kwargs)
    else:
        return GeoObject(*args)



import functools

from .logger import logger
from .misc import normalize_path


class PathTraverser:
    MAX_SLICE_LENGTH = 10

    def __init__(self, path, *args, **kwargs):
        self._path = normalize_path(path)

    @property
    def is_multiple(self):
        return self._path != None and len(self._path) > 0 and functools.reduce(lambda a, b: a and b, [isinstance(p, slice) for p in self._path], True)

    def append(self, seg):
        self._path.append(seg)

    @property
    def path(self):
        return self._path

    def traverse(self,  visitor, prev=[], idx=0):
        res = None
        if self._path is None or idx >= len(self._path):
            res = visitor(prev)
        elif isinstance(self._path[idx], str) or isinstance(self._path[idx], int):
            res = self.traverse(visitor, prev+[self._path[idx]], idx+1)
        elif isinstance(self._path[idx], slice):
            start = self._path[idx].start or 0
            step = self._path[idx].step or 1
            stop = self._path[idx].stop
            if stop is not None:
                res = [self.traverse(visitor, prev+[sidx], idx+1) for sidx in range(start,  stop, step)]
            else:
                res = []
                count = 0
                while count < PathTraverser.MAX_SLICE_LENGTH:
                    try:
                        res.append(self.traverse(visitor, prev+[start], idx+1))
                    except StopIteration:
                        break
                    except IndexError:
                        break
                    count = count + 1
                    start = start+step
                if count >= PathTraverser.MAX_SLICE_LENGTH:
                    logger.warning("List length > {PathTraverser.MAX_SLICE_LENGTH}")

        return res

    def apply(self, visitor):
        res = self.traverse(visitor, [], 0)
        return res[0] if self.is_multiple else res

    def iter_traverse(self, prev=[], idx=0):
        if idx >= len(self._path):
            yield prev
        elif isinstance(self._path[idx], str) or isinstance(self._path[idx], int):
            yield from self.iter_traverse(prev+[self._path[idx]], idx+1)
        elif isinstance(self._path[idx], slice):
            start = self._path[idx].start or 0
            step = self._path[idx].step or 1
            stop = self._path[idx].stop
            if stop is not None:
                for sidx in range(start,  stop, step):
                    yield from self.iter_traverse(prev+[sidx], idx+1)
            else:
                res = []
                count = 0
                while count < PathTraverser.MAX_SLICE_LENGTH:
                    try:
                        yield from self.iter_traverse(prev+[start], idx+1)
                    except StopIteration:
                        break
                    except IndexError:
                        break
                    count = count + 1
                    start = start+step
                if count >= PathTraverser.MAX_SLICE_LENGTH:
                    logger.warning(f"Index > MAX_SLICE_LENGTH({PathTraverser.MAX_SLICE_LENGTH})! {prev+[start]}")

    def __iter__(self):
        yield from self.iter_traverse()



from __future__ import annotations

import collections.abc
import pathlib
import typing


from ..utils.logger import logger
from ..utils.uri_utils import URITuple, uri_split
from .Document import Document
from .Entry import Entry


class File(Document):
    """
    File like object
    """

    def __init__(self, url: str | pathlib.Path | URITuple, *args, format=None, default_format=None, **kwargs):
        if self.__class__ is File:
            if format is not None:
                format = format.lower()
            elif isinstance(url, dict):
                format = url.get("$class", "").lower()
            elif isinstance(url, pathlib.PosixPath):
                format = url.suffix[1:].lower()
            elif isinstance(url, (str, URITuple)):
                uri = uri_split(url)
                schemes = uri.protocol.split("+") if uri.protocol != "" and uri.protocol != None else []
                if len(schemes) == 0:
                    pass
                elif schemes[0] in ["file", "local"]:
                    format = "+".join(schemes[1:])

                if format is None or format == "":
                    format = pathlib.PosixPath(uri.path).suffix[1:]

            super().__dispatch_init__([format, default_format], self, url, *args, **kwargs)

            return

        super().__init__(url, *args, **kwargs)
        self._is_open = False

    def __del__(self):
        fid = getattr(self, "_fid", None)
        if fid is not None:
            fid.close()
            self._fid = None

    @property
    def mode_str(self) -> str:
        return File.MOD_MAP.get(self.mode, "r")

    @property
    def is_writable(self) -> bool:
        return self.mode | File.Mode.write > 0

    # @property
    # def entry(self) -> Entry: return FileEntry(file=self)

    def __enter__(self) -> Document:
        return super().__enter__()

    def read(self, lazy=False) -> Entry:
        raise NotImplementedError(f"TODO: {self.__class__.__name__}.read")

    def write(self, data=None, *args, lazy=False, **kwargs):
        raise NotImplementedError(f"TODO: {self.__class__.__name__}.write")


class FileEntry(Entry):
    def __init__(self, *args, file, **kwargs):
        super().__init__(*args, **kwargs)
        self._fid = file

    def __copy_from__(self, other: FileEntry) -> Entry:
        super().__copy_from__(other)
        self._fid = other._fid
        return self

    # def __del__(self):
    #     if len(self._path) == 0:
    #         self.flush()

    def flush(self):
        logger.debug(self._data)
        self._fid.write(self._data)
        self._data = None



import unittest

import numpy as np
import scipy.constants

from spdm.core.Expression import Variable
from spdm.core.Field import Field
from spdm.utils.logger import logger

TWOPI = scipy.constants.pi * 2.0


class TestField(unittest.TestCase):
    def test_attribute(self):
        x = np.linspace(0, 1 * TWOPI, 128)
        y = np.linspace(0, 2 * TWOPI, 128)

        _x = Variable(0, "x")
        _y = Variable(1, "y")
        fun = Field(x, y, np.sin(_x) * np.cos(_y), mesh_periods=[TWOPI, 2 * TWOPI])

        self.assertEqual(fun.mesh.ndim, 2)
        self.assertTrue(np.allclose(fun.mesh.dims[0], x))
        self.assertTrue(np.allclose(fun.mesh.dims[1], y))

    def test_spl2d(self):
        _x = Variable(0, "x")
        _y = Variable(1, "y")
        x = np.linspace(0, 1 * TWOPI, 128)
        y = np.linspace(0, 2 * TWOPI, 128)
        g_x, g_y = np.meshgrid(x, y)

        z = np.sin(g_x) * np.cos(g_y)

        fun = Field(x, y, np.sin(_x) * np.cos(_y), mesh_periods=[TWOPI, 2 * TWOPI])

        z2 = fun(g_x, g_y)

        self.assertTrue(np.allclose(z, z2, rtol=1.0e-4))

    def test_pd(self):
        _x = Variable(0, "x")
        _y = Variable(1, "y")
        x = np.linspace(0, TWOPI, 128)
        y = np.linspace(0, 2 * TWOPI, 128)
        g_x, g_y = np.meshgrid(x, y)

        Z = Field(x, y, np.sin(_x) * np.cos(_y), mesh_periods=[TWOPI, 2 * TWOPI])

        self.assertTrue(np.allclose(np.sin(g_x) * np.cos(g_y), Z(g_x, g_y), rtol=1.0e-4))

        self.assertTrue(np.allclose(np.cos(g_x) * np.cos(g_y), Z.pd(1, 0)(g_x, g_y), rtol=1.0e-4))

        # dzdx = TWOPI*np.cos(g_x*TWOPI)*np.cos(g_y*TWOPI)
        # dZdx = Z.pd(1, 0)(g_x, g_y)
        # # logger.debug(np.count_nonzero(~np.isclose(dzdx, dZdx, rtol=1.0e-4)))
        # self.assertTrue(np.allclose(dzdx, dZdx, rtol=1.0e-4))

        # dzdy = -TWOPI*np.sin(g_x*TWOPI)*np.sin(g_y*TWOPI)
        # dZdy = Z.pd(0, 1)(g_x, g_y)
        # logger.debug(np.count_nonzero(~np.isclose(dzdy, dZdy, rtol=1.0e-4)))
        # self.assertTrue(np.allclose(dzdy, dZdy, rtol=1.0e-4))

        # dZdxdy = Z.pd(1, 1)(g_x, g_y)
        # dzdxdy = -(TWOPI)*(TWOPI)*np.cos(g_x*TWOPI)*np.sin(g_y*TWOPI)
        # logger.debug(np.count_nonzero(~np.isclose(dzdxdy, dZdxdy, rtol=1.0e-4)))
        # self.assertTrue(np.allclose(dzdxdy, dZdxdy, rtol=1.0e-4))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from scipy import constants
from spdm.core.Expression import Expression, Piecewise, Variable
from spdm.utils.logger import logger
from spdm.core.Domain import Domain

TWOPI = constants.pi*2.0


class TestExpression(unittest.TestCase):
    def test_define(self):

        _x = Variable(0, "x")
        _y = Variable(1, "y")

        d = Domain(_x > 0.3, _x < 0.7)

        self.assertTrue(d(0.5))
        self.assertFalse(d(0.3))

        x = np.linspace(0, 1, 10)
        y = np.zeros(10)
   
        y[d(x)] = 1
        y_ = np.asarray([0., 0., 0., 1., 1., 1., 1., 0., 0., 0.])
        self.assertTrue(np.allclose(y, y_))

    def test_operation(self):

        _x = Variable(0, "x")

        d1 = Domain(_x > 0.3)
        d2 = Domain(_x < 0.7)

        d3 = ~(d1 & d2)

        x = np.linspace(0, 1, 10)
        y = np.zeros(10)
    
        y[d3(x)] = 1
        y_ = np.asarray([1., 1., 1., 0., 0., 0., 0., 1., 1., 1.])
        self.assertTrue(np.allclose(y, y_))


if __name__ == '__main__':
    unittest.main()



from copy import copy, deepcopy

import numpy as np
from spdm.utils.logger import logger
from spdm.core.Dict import Dict
from spdm.core.HTree import HTree
from spdm.core.sp_property import sp_property
from spdm.core.List import List


# class Foo(Dict[Node]):
#     def __init__(self, *args, **kwargs) -> None:
#         super().__init__(*args, **kwargs)
#     foo: str = sp_property()  # type: ignore
#     doo: int = sp_property(default_value=10)  # type: ignore

#     @sp_property
#     def goo(self) -> int:
#         return self.doo+100

class Ion(Dict):
    label: str = sp_property()  # type:ignore
    Z: int = sp_property()  # type:ignore
    m: float = sp_property()  # type:ignore
    density: float = sp_property()  # type:ignore


class TransModel(Dict):
    label: str = sp_property()  # type:ignore
    ions: List[Ion] = sp_property()  # type:ignore


if __name__ == '__main__':

    atom_list = List[TransModel]([
        {"id": "first",
            "ions":  [
                {"label": "H", "Z": 1, "m": 1, "density": 1.0},
                {"label": "D", "Z": 1, "m": 3},
                {"label": "T", "Z": 1, "m": 3},
            ]},
        {"id": "second",
            "ions":  [
                {"label": "H", "Z": 1, "m": 1, "density": 2.0},
                {"label": "D", "Z": 1, "m": 3},
                {"label": "T", "Z": 1, "m": 3},
            ]}

    ])

    logger.debug(atom_list[0].ions[0].label)
    # ion = atom_list.combine(common_data={"id": "commbine"})

    # for d in ion.ions.find({"label": "H"}):
    #     logger.debug(d)

    # logger.debug(ion.ions[{"label": "H", "@only_first": False}].density)

    # logger.debug(ion.Z)
    # logger.debug(ion.density)

    # cache = {
    #     "a": [
    #         "hello world {name}!",
    #         "hello world2 {name}!",
    #         1.0, 2, 3, 4
    #     ],
    #     "c": "I'm {age}!",
    #     "d": {
    #         "e": "{name} is {age}",
    #         "foo": "{address}"
    #     }
    # }
    # d = Dict[Foo](cache)
    # logger.debug(d["d"])
    # tmp = d["d"]
    # logger.debug(tmp["foo"])
    # logger.debug(tmp.foo)
    # logger.debug(tmp["foo"])
    # logger.debug(tmp.foo)
    # logger.debug(tmp["doo"])
    # logger.debug(tmp.doo)
    # logger.debug(tmp["doo"])
    # logger.debug(tmp.goo)
    # logger.debug(tmp["goo"])
    # # d["b"] = "hello world {name}!"

    # # d["h"].append(1234)
    # # d["h"].append(5678)

    # # d.update({"d": {"g": 5}})

    # # print(cache)






def intersection(self, other: GeoObject) -> typing.List[GeoObject]:
        """ Return the intersection of self with other. """
        return [bbox_intersection(self.bbox, other.bbox)]

def reflect(self, line: GeoObject) -> GeoObject: return bbox_reflect(self.bbox, line)
""" reflect  by line"""

def rotate(self, angle, axis=None) -> GeoObject: return bbox_rotate(self.bbox, angle, axis=axis)
""" rotate  by angle and axis"""

def scale(self, *s, point=None) -> GeoObject: return bbox_scale(self.bbox, *s, point=point)
""" scale self by *s, point """

def translate(self, *shift) -> GeoObject: return bbox_translate(self.bbox, *shift)

def trim(self): raise NotImplementedError(f"")


import unittest
from copy import deepcopy
import pathlib
from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_
from spdm.core.Entry import Entry, as_value
from spdm.core.File import File
xml_file = pathlib.Path(__file__).parent.joinpath("pf_active.xml")


class TestXMLEntry(unittest.TestCase):

    def test_read(self):
        entry = File(xml_file).read()

        self.assertEqual(entry.child("coil/0/name").__value__, "PF1")

    def test_iter(self):

        entry = File(xml_file).read()

        name_list = ([v.child('name').__value__ for v in entry.child("coil")])

        name_list_expect = ['PF1', 'PF2', 'PF3', 'PF4', 'PF5', 'PF6', 'PF7', 'PF8',
                            'PF9', 'PF10', 'PF11', 'PF12', 'PF13', 'PF14', 'IC1', 'IC2']

        self.assertListEqual(name_list, name_list_expect)

    def test_exists(self):
        entry = File(xml_file).read()

        self.assertTrue(entry.child("coil/0/name").exists)
        self.assertFalse(entry.child("coil/0/key").exists)

    def test_count(self):
        entry = File(xml_file).read()
        self.assertEqual(entry.child("coil").count, 16)


if __name__ == '__main__':
    unittest.main()



from .logger import logger
import subprocess

def subprocess(cmd, working_dir='.', **kwargs):
    # logger.debug(f"CMD: {cmd} : {res}")
    logger.info(f"Execute Shell command [{working_dir}$ {' '.join(cmd)}]")
    # @ref: https://stackoverflow.com/questions/21953835/run-subprocess-and-print-output-to-logging
    try:
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            # env=self.envs,
            shell=True,
            cwd=working_dir,
            capture_output=True,
            check=False,
            **kwargs
        )
        # process_output, _ = command_line_process.communicate()

        with process.stdout as pipe:
            for line in iter(pipe.readline, b''):  # b'\n'-separated lines
                logger.info(line)

        exitcode = process.wait()

    except (OSError, subprocess.CalledProcessError) as error:
        logger.error(
            f"""Command failed! [{cmd}]
                   STDOUT:[{error.stdout}]
                   STDERR:[{error.stderr}]""")
        raise error
    
    return exitcode



import functools
from spdm.core.AttributeTree import attribute
from ..util.logger import logger


class Doo:
    def __init__(self, v, *args, **kwargs) -> None:
        self._v = v

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._v} />"


class Foo:

    @attribute
    def attr1(self):
        r"Attribute One"
        return {"a": 1.0}

    @attribute(Doo)
    def attr2(self):
        r"Attribute One"
        return {"a": 2.0}

    @functools.cached_property
    def attr3(self):
        return 1.0


if __name__ == "__main__":
    foo = Foo()
    logger.debug(foo.attr1)
    logger.debug(foo.attr2)
    foo.attr3 = {"b": 1.234}
    logger.debug(foo.attr3)



import collections
import typing
import numpy
import yaml
from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.utils.logger import logger
from spdm.utils.typing import as_native


# class YAMLDocument(Document):
#     def __init__(self, path, *args, **kwargs):
#         super().__init__(path, *args, **kwargs)
#         self.load(path, mode=self.mode)

#     def load(self, *args, **kwargs):
#         with self.open(mode="r") as fid:
#             if fid is not None:
#                 self._holder = yaml.load(fid, Loader=yaml.CLoader)
#             else:
#                 self._holder = None

#     def save(self, d, *args, **kwargs):
#         with self.open(mode="w") as fid:
#             yaml.dump(self._holder, fid,  Dumper=yaml.CDumper)


# class YAMLCollection(Collection):
#     def __init__(self, uri, *args, **kwargs):
#         super().__init__(uri, *args,
#                          file_extension=".json",
#                          file_factory=lambda *a, **k: YAMLDocument(*a, **k),
#                          ** kwargs)


@File.register(["yaml", "YAML"])
class FILEPLUGINyaml(File):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fid: typing.Optional[typing.IO[typing.Any]] = None

    def open(self) -> File:
        super().open()
        try:
            self._fid = open(self.path,  mode=self.mode_str)
        except OSError as error:
            raise FileExistsError(f"Can not open file {self.path}! {error}")
        else:
            logger.debug(
                f"Open {self.__class__.__name__} File {self.path} mode={self.mode}")
        return self

    def close(self):
        if self._fid is not None:
            self._fid.close()
            self._fid = None
        return super().close()

    def read(self, *args,   **kwargs) -> Entry:
        if not hasattr(self, "_fid"):
            self.open()
        return Entry(yaml.load(self._fid, Loader=yaml.CLoader))

    def write(self,   d, *args,  **kwargs):
        yaml.dump(as_native(d, enable_ndarray=False), self._fid,  Dumper=yaml.CDumper)


__SP_EXPORT__ = FILEPLUGINyaml



from ..utils.typing import array_type
from .Function import Function
from .sp_property import sp_tree, sp_property


@sp_tree
class Signal:
    """Signal with its time base"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._func = None

    @property
    def name(self) -> str:
        return self._metadata.get("name", "")

    data: array_type

    time: array_type = sp_property(units="s")

    def __call__(self, t: float) -> float:
        if self._func is None:
            self._func = Function(self.time, self.data)
        return self._func(t)


class SignalND(Signal):
    pass



import typing

from .HTree import Dict, HTree, List, HTreeNode
from .Edge import Edge
from .Path import PathLike

_T = typing.TypeVar("_T")


class Graph(Dict[_T]):
    """Represents '''Graph'''.
    * defines namespace for the '''Node'''s
    * Graph is a Node

    TODO (salmon 2019.7.25): add subgraph
    """

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(value, *args, **kwargs)
        self._edges: typing.List[Edge] = []

    @property
    def edges(self) -> typing.List[Edge]:
        return self._edges

    def link(self, source: PathLike, target: PathLike, *args, **kwargs) -> Edge:
        e = Edge(source, target, *args, graph=self, **kwargs)
        self._edges.append(e)
        return e


# class FunctionWrapperGraph(Graph):
#     def __init__(self, func, *args, **kwargs):
#         super().__init__(*args, signature=inspect.signature(func), **kwargs)


# def as_graph(func=None, **attrs):
#     def _decorate(wrapped):
#         @functools.wraps(wrapped)
#         def _wrapper(*args, **kwargs):
#             obj = None
#             if inspect.isfunction(wrapped):
#                 obj = FunctionWrapperActor(wrapped,  **attrs)
#             else:
#                 raise ValueError(f"Can not convert {wrapped} to Actor!")
#             obj.bind(*args, **kwargs)
#             return obj
#         return _wrapper

#     if func is None:
#         return _decorate
#     else:
#         return _decorate(func)



import pprint
import subprocess
import pathlib
import numpy as np

from spdm.data import io
from ..util.logger import logger

if __name__ == "__main__":

    # subprocess.run(['mongod',
    #                 '--dbpath', pathlib.Path.cwd(),
    #                 '--logpath', pathlib.Path.cwd()/"mongodb.log",
    #                 '--bind_ip', '127.0.0.1'
    #                 ])

    c1 = io.connect("mongodb://127.0.0.1:27017/EAST/diag/imas")
    # collection.insert_one({"a": "Just a test", "b": 5})
    # c1 = io.connect("mdsplus://127.0.0.1/EAST")

    for i in range(10):
        c1.insert_one(
            {"device": {"name": "EAST", "shot": 555}, "num": i,
             "data": {
                 "x": np.random.rand(5, 6),
                "y": np.random.rand(5, 6),
            }
            })

    pprint.pprint([d["data"] for d in c1.find()])
    # pprint.pprint([d for d in c1.find(projection="data")])



import typing
import abc
from .GeoObject import GeoObject
from .Plane import Plane


class Solid(GeoObject):
    """ Line
        线，一维几何体
    """

    def __init__(self, *args,  **kwargs) -> None:
        super().__init__(*args, rank=3, **kwargs)

    @abc.abstractproperty
    def points(self, *args, **kwargs):
        raise NotImplementedError(f"{self.__class__.__name__}")

    @abc.abstractproperty
    def boundary(self) -> typing.List[Plane]:
        raise NotImplementedError(f"{self.__class__.__name__}")



from __future__ import annotations

import typing

import numpy as np
from scipy.interpolate import CubicSpline, PPoly

from .Mesh import Mesh
from .RectilinearMesh import RectilinearMesh


@Mesh.register('ppoly')
class PPolyMesh(RectilinearMesh):
    """Piecewise polynomial Mesh
    """

    def interpolator(self, y, **kwargs) -> PPoly:
        bc_type = self._metadata.get("bc_type", "not-a-knot")
        #  "periodic"         "periodic" if np.all(y[0] == y[-1]) else
        # return interp1d(*self.points, y, **kwargs)
        return CubicSpline(*self.xyz, y, bc_type=bc_type, **kwargs)


__SP_EXPORT__ = PPolyMesh

# class Function1D(Function[_T]):
#     """
#
#     """
#     def __init__(self, x, y,  **kwargs):
#         super().__init__(x, y, **kwargs)

#         if len(self._Mesh) == 0:
#             raise RuntimeError(f"Function must have at least one 'x' argument!")

#         # if len(args) == 0:
#         #     x = kwargs.get("x", 0)
#         #     y = kwargs.get("y", 0)
#         # elif len(args) == 1:
#         #     y = args[0]
#         #     x = None
#         # else:
#         #     x = args[0]
#         #     y = args[1]
#         #     if len(args) > 2:
#         #         raise RuntimeWarning(f"Too much position arguments {args}")
#         # if isinstance(self._x, np.ndarray):
#         #     assert(self._x[0] < self._x[-1])
#         # if y is None or (isinstance(y, (np.ndarray, collections.abc.Sequence)) and len(y) == 0):
#         #     self._data = 0

#         # if isinstance(y, Function):
#         #     if x is None:
#         #         x = y.x_domain
#         #         self._data = y._data
#         #     else:
#         #         self._data = y
#         # elif isinstance(y, PPoly):
#         #     self._data = y
#         #     if x is None:
#         #         x = y.x
#         # else:
#         #     self._data = y

#         # if x is None or len(x) == 0:
#         #     self._x = None
#         #     self._x_domain = [-np.inf, np.inf]
#         # elif isinstance(x, np.ndarray):
#         #     if len(x) == 0:
#         #         logger.error(f"{type(x)} {type(y)}")
#         #     self._x = x
#         #     self._x_domain = [x[0], x[-1]]
#         # elif isinstance(x, collections.abc.Sequence) and len(x) > 0:
#         #     self._x_domain = list(set(x))
#         #     if isinstance(y, np.ndarray):
#         #         self._x = np.linspace(
#         #             self._x_domain[0], self._x_domain[-1], len(y))
#         #     else:
#         #         self._x = None
#         # else:
#         #     self._x_domain = [-np.inf, np.inf]
#         #     self._x = None

#         # if self._data is None or self._data is _not_found_:
#         #     logger.warning(f"Empty function: x={self._x},y={self._data}")
#         # elif isinstance(self._data, np.ndarray) and (self._x is not None and self._x.shape != self._data.shape):
#         #     raise ValueError(f"x.shape  != y.shape {self._x.shape}!={self._data.shape}")
#     def __str__(self) -> str:
#         return pprint.pformat(self.__array__())
#     @property
#     def is_valid(self) -> bool:
#         return self._Mesh is not None and self._data is not None
#     @cached_property
#     def is_constant(self) -> bool:
#         return isinstance(self._data, (int, float))

# @cached_property
# def is_periodic(self) -> bool:
#     return self.is_constant \
#         or (isinstance(self._data, np.ndarray) and np.all(self._data[0] == self._data[-1])) \
#         or np.all(self.__call__(self.x_min) == self.__call__(self.x_max))
# @property
# def is_bounded(self):
#     return not (self.x_min == -np.inf or self.x_max == np.inf)
# @property
# def continuous(self) -> bool:
#     return len(self.x_domain) == 2
# @property
# def x_min(self) -> float:
#     return self.x_domain[0]
# @property
# def x_max(self) -> float:
#     return self.x_domain[-1]
# @property
# def _Mesh(self) -> np.ndarray:
#     return self._x  # type:ignore
# @_Mesh.setter
# def _Mesh(self, x: np.ndarray):
#     if isinstance(self._data, np.ndarray) and len(self._data) != len(x):
#         raise ValueError(f"len(x) != len(y) {len(x)} != {len(self._data)}")
#     self._x = x
# def setdefault_x(self, x):
#     if self._x is None:
#         self._Mesh = x
#     return self._x
# def __len__(self) -> int:
#     if self._Mesh is not None:
#         return len(self._Mesh)
#     else:
#         raise RuntimeError(f"Can not get length from {type(self._data)} or {type(self._Mesh)}")
# def __array__(self) -> np.ndarray:
#     if isinstance(self._data, np.ndarray):
#         return self._data
#     elif hasattr(self._data.__class__, "__entry__"):
#         v = self._data.__entry__().__value__()
#         if v is None or v is _not_found_:
#             raise ValueError(f"Can not get value from {self._data}")
#         self._data = np.asarray(v, dtype=self.__type_hint__)
#     else:
#         return self.__call__()
# @cached_property
# def __fun_op__(self) -> PPoly:
#     if isinstance(self._data,  PPoly):
#         return self._data
#     elif self._Mesh is None:
#         raise ValueError(f"_Mesh is None")
#     elif isinstance(self._data, np.ndarray):
#         assert (self._Mesh.size == self._data.size)
#         return create_spline(self._Mesh,  self._data)
#     else:
#         return create_spline(self._Mesh,  self.__call__(self._Mesh))
# def __call__(self, x=None, /,  **kwargs) -> typing.Union[np.ndarray, float]:
#     if x is None:
#         x = self._Mesh
#     if x is None:
#         raise RuntimeError(f"_Mesh is None!")
#     if x is self._Mesh and isinstance(self._data, np.ndarray):
#         return self._data
#     if self._data is None:
#         raise RuntimeError(f"Illegal function! y is None {self.__class__}")
#     elif isinstance(self._data, (int, float)):
#         if isinstance(x, np.ndarray):
#             return np.full(x.shape, self._data)
#         else:
#             return self._data
#     elif callable(self._data):
#         return np.asarray(self._data(x, **kwargs), dtype=float)
#     # elif hasattr(y, "__array__"):
#     #     y = y.__array__
#     elif x is not self._Mesh and isinstance(self._data, np.ndarray):
#         return self.__fun_op__(x, **kwargs)
#     else:
#         raise TypeError((type(x), (self._data)))
# def resample(self, x_min, x_max=None, /, **kwargs):
#     if x_min is None or (x_max is not None and x_min <= self.x_min and self.x_max <= x_max):
#         if len(kwargs) > 0:
#             logger.warning(f"ignore key-value arguments {kwargs.keys()}")
#             # TODO: Insert points in rapidly changing places.
#         return self
#     elif x_max is None:
#         return Function(x_min, self.__call__(x_min, **kwargs))
#     x_min = max(self.x_min, x_min)
#     x_max = min(self.x_max, x_max)
#     if x_min > x_max or np.isclose(x_min, x_max) or x_max <= self.x_min:
#         raise ValueError(f"{x_min,x_max}  not in  {self.x_min,self.x_max}")
#     elif isinstance(self._Mesh, np.ndarray):
#         idx_min = np.argmax(self._Mesh >= x_min)
#         idx_max = np.argmax(self._Mesh > x_max)
#         if idx_max > idx_min:
#             pass
#         elif idx_max == 0 and np.isclose(self._Mesh[-1], x_max):
#             idx_max = -1
#         else:
#             logger.debug((x_min, x_max, idx_min, idx_max, self._Mesh))
#         if isinstance(self._data, np.ndarray):
#             return Function(self._Mesh[idx_min:idx_max], self._data[idx_min:idx_max])
#         else:
#             return Function(self._Mesh[idx_min:idx_max],  self.__call__(self._Mesh[idx_min:idx_max]))
#     elif callable(self._data):
#         return Function([x_min, x_max], self._data)
#     else:
#         raise TypeError((type(self._Mesh), type(self._data)))
#         # return _Mesh, np.asarray(self.__call__(_Mesh), dtype=float)
# def __setitem__(self, idx, value):
#     if hasattr(self, "_interpolator"):
#         delattr(self, "_interpolator")
#     self.__real_array__()[idx] = value
# def _prepare(self, x, y):
#     if not isinstance(x, [collections.abc.Sequence, np.ndarray]):
#         x = np.asarray([x])
#     if isinstance(y, [collections.abc.Sequence, np.ndarray]):
#         y = np.asarray(y)
#     elif y is not None:
#         y = np.asarray([y])
#     else:
#         y = self.__call__(x)
#     return x, y
# def insert(self, x, y=None):
#     res = Function(*self._prepare(x, y), func=self._func)
#     raise NotImplementedError('Insert points!')
#     return res
# def __len__(self):
#     return len(self.x) if self.x is not None else 0



import unittest
from copy import deepcopy
import pprint
from spdm.core.Path import Path
from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_


class TestPath(unittest.TestCase):
    def test_init(self):
        p = ["a", "b", {"c": slice(4, 10, 2)}]
        self.assertEqual(Path(p)[:], p)

    def test_parser(self):
        self.assertEqual(Path._parser("a/b/c"), ["a", "b", "c"])
        self.assertEqual(Path._parser("a/b/c/0"), ["a", "b", "c", 0])
        self.assertEqual(Path._parser("a/b/c/0/d"), ["a", "b", "c", 0, "d"])

        self.assertEqual(Path._parser("a[(1,2,3,'a')]/h"), ["a", (1, 2, 3, "a"), "h"])
        self.assertEqual(Path._parser("a[{1,2,3,'a'}]/h"), ["a", {1, 2, 3, "a"}, "h"])
        # self.assertEqual(Path._parser("a[{'$le':[1,2]}]/h"), ["a", {Path.tags.le: [1, 2]}, "h"])
        # self.assertEqual(Path._parser("a[1:10:-3]/h"),       ["a", slice(1, 10, -3), "h"])
        # self.assertEqual(Path._parser("a[1:10:-3]/$next"),   ["a", slice(1, 10, -3), Path.tags.next])

    def test_append(self):
        p = Path()
        p.append("a/b/c")

        self.assertEqual(p[:], ["a", "b", "c"])

    data = {
        "a": ["hello world {name}!", "hello world2 {name}!", 1, 2, 3, 4],
        "c": "I'm {age}!",
        "d": {"e": "{name} is {age}", "f": "address"},
    }

    def test_query(self):
        self.assertEqual(Path("c").find(self.data), self.data["c"])
        self.assertEqual(Path("d/e").find(self.data), self.data["d"]["e"])
        self.assertEqual(Path("d/f").find(self.data), self.data["d"]["f"])
        self.assertEqual(Path("a/0").find(self.data), self.data["a"][0])
        self.assertEqual(Path(["a", 1]).find(self.data), self.data["a"][1])
        # self.assertEqual(Path("a/2:4:1").query(self.data),        self.data["a"][2:4])
        self.assertEqual(Path("d/k").find(self.data, default_value=None), None)

    def test_query_op(self):
        self.assertEqual(Path("a").find(self.data, Path.tags.count), 6)
        self.assertEqual(Path("d/e").find(self.data, Path.tags.count), 1)
        self.assertEqual(Path("b/h").find(self.data, Path.tags.count), 0)
        self.assertEqual(Path("d/f").find(self.data, Path.tags.count), 1)

        self.assertEqual(Path().find(self.data, Path.tags.count), 3)
        self.assertEqual(Path("a").find(self.data, Path.tags.count), 6)
        self.assertEqual(Path("d").find(self.data, Path.tags.count), 2)

        # self.assertTrue(Path(["a", slice(2, 7), {Path.tags.equal: [1, 2, 3, 4]}]).query(self.data))

    def test_get_many(self):
        cache = deepcopy(self.data)

        res = Path({"a/2", "c", "d/e", "e"}).find(cache)

        self.assertDictEqual(res, {"a/2": cache["a"][2], "c": cache["c"], "d/e": cache["d"]["e"], "e": _not_found_})

    def test_insert(self):
        cache = {}

        Path("a").insert(cache, "hello world {name}!")
        self.assertEqual(cache["a"], "hello world {name}!")

        Path("e/f").insert(cache, 5)
        Path("e/g").insert(cache, 6)

        self.assertEqual(cache["e"]["f"], 5)
        self.assertEqual(cache["e"]["g"], 6)

    def test_insert_ext(self):
        cache = deepcopy(self.data)

        Path("c").insert(cache, {"a": "hello world", "b": 3.141567})
        Path("c").insert(cache, 1.23455)

        self.assertEqual(cache["c"][0], "I'm {age}!")
        # self.assertEqual(cache["c"][1]["a"], "hello world")
        # self.assertEqual(cache["c"][1]["b"], 3.141567)
        self.assertEqual(cache["c"][-1], 1.23455)

    def test_update(self):
        cache = deepcopy(self.data)

        Path().update(cache, {"d": {"g": 5, "f": 6}})

        self.assertEqual(cache["d"]["e"], "{name} is {age}")

        self.assertEqual(cache["d"]["f"], 6)

        self.assertEqual(cache["d"]["g"], 5)

    def test_delete(self):
        cache = {"a": ["hello world {name}!", "hello world2 {name}!", 1, 2, 3, 4], "b": "hello world!"}

        Path("b").remove(cache)

        self.assertTrue(Path("b").find(cache) is _not_found_)

    def test_sequence_tag(self):
        cache = {
            "people": [
                {"@name": "zhangsan", "age": 10, "address": "beijing"},
                {"@name": "lisi", "age": 20, "address": "shanghai"},
                {"@name": "wangwu", "age": 30, "address": "guangzhou"},
                {"@name": "zhaoliu", "age": 40, "address": "shenzhen"},
            ]
        }

        self.assertEqual(Path("people/zhangsan/age").get(cache), 10)
        self.assertEqual(Path("people/lisi/age").get(cache), 20)

        Path("people/qianqi/age").put(cache, 70)
        self.assertEqual(Path("people/qianqi/age").get(cache), 70)

    # def test_find_many(self):

    #     res = Path(("a/2", "c",  "d/e", "e")).query(self.data, default_value=_not_found_)

    #     self.assertListEqual(res, [self.data['a'][2],
    #                                self.data['c'],
    #                                self.data['d']['e'],
    #                                _not_found_])


if __name__ == "__main__":
    unittest.main()




import sys

import matplotlib.pyplot as plt
import numpy as np
from scipy import constants
from spdm.utils.plot_profiles import plot_profiles
from spdm.data import Function
from ..util.logger import logger

if __name__ == '__main__':

    x = np.linspace(0, constants.pi*2.0, 64, endpoint=True)
    y = np.sin(x)
    m = np.zeros([3, 64])
    # f0 = Function(x, y, is_periodic=True)
    # f1 = Function(np.linspace(0, 1.0, 128), y, is_periodic=True)
    f1 = Function(x, [(lambda t:t < 1), (lambda t:t >= 1)], [lambda t:t**2, lambda t:0.4*np.sin(t)])
    # f2 = (f1*2+x)*m
    # logger.debug(f2.shape)
    x2 = np.linspace(0, constants.pi*2.0, 128, endpoint=True)
    f3 = Function(x, np.sin)

    d = (f3*2+f1*x)(x2)

    plot_profiles([
        # [(f0, r"$f0(x)$"),
        #  (f1, r"$f1(x)=f0(2\pi x)$")],
        # [(np.cos(x), r"$cos(x)$"),
        #  (f0.derivative, r"$d f0(x)$"),
        #  (f1.derivative/(constants.pi*2.0), r"$d f1(x)/2\pi$")],
        # [(-np.cos(x)+np.cos(0), r"-$cos(x)+cos(0)$"),
        #  (f0.antiderivative, r"$\int f0$"),
        #  (constants.pi*2.0*f1.antiderivative, r"2\pi $\int f1$")],
        (f1,  "f1"),
        # (f2,  "f2"),
        (f3,  "f3")
    ],
        x_axis=(x2, "x")
    ).savefig("/home/salmon/workspace/output/test_function.svg")



import typing

from numpy.typing import ArrayLike, NDArray

from ..core.Container import Container
from ..core.Function import Function
from .Mesh import Mesh


class MultiBlockMesh(Container[Mesh]):

    def __init__(self, block_list: typing.List[typing.Any],  **kwargs) -> None:
        super().__init__(block_list, **kwargs)

        self._cond_list = cond_list
        self._Mesh_list: typing.List[Mesh] = Mesh_list

    def interpolator(self, *args, **kwargs) -> Function:
        return self._cond_list


class MultiBlockFunction(Function):

    def __init__(self, block_list: typing.List[typing.Any], **kwargs) -> None:
        super().__init__(**kwargs)

        self._Mesh_list: MultiBlockMesh = cond_list
        self._data_list: typing.List[Mesh] = Mesh_list

    def __call__(self, *args, **kwargs) -> ArrayLike | NDArray:

        return super().__call__(*args, **kwargs)




import collections.abc
import typing
from functools import cached_property

import numpy as np

from ..utils.logger import logger
from ..utils.misc import builtin_types
from ..utils.plugin import Pluggable
from ..utils.typing import (ArrayLike, ArrayType, NumericType, ScalarType,
                            nTupleType, numeric_type)
from .GeoObject import GeoObject


class ParametricGeoObj(GeoObject):
    """
        参数化几何对象（Parametric geometry object）是指使用参数方程来定义的几何对象。
        参数方程是一种数学方法，它定义了一组量作为一个或多个独立变量（称为参数）的函数1。
        参数方程通常用于表示构成几何对象（如曲线或曲面）的点的坐标，分别称为参数曲线和参数曲面1。
    """
    def remesh(self, *args, **kwargs): raise NotImplementedError(f"")

    def coordinates(self, *uvw) -> NumericType:
        """
            将 _参数坐标_ 转换为 _空间坐标_
            @return: array-like shape = [*uvw.shape[:-1],ndim]
        """
        raise NotImplementedError(f"{self.__class__.__name__}")

    def parametric_coordinates(self, *xyz) -> NumericType:
        """
            将 _空间坐标_ 转换为 _参数坐标_         
            @return: array-like shape = [*uvw.shape[:-1],rank]
        """
        raise NotImplementedError(f"{self.__class__.__name__}.parametric_coordinates")

    def xyz(self, *uvw) -> NumericType: return self.coordinates(*uvw)
    """ alias of coordinates """

    def uvw(self, *xyz) -> NumericType: return self.parametric_coordinates(*xyz)
    """ alias of parametric_coordinates """

    def dl(self, uv=None) -> NumericType: return np.asarray(0)
    """
        derivative of shape
        Returns:
            rank==0 : 0
            rank==1 : dl (shape=[n-1])
            rank==2 : dx (shape=[n-1,m-1]), dy (shape=[n-1,m-1])
            rank==3 : dx (shape=[n-1,m-1,l-1]), dy (shape=[n-1,m-1,l-1]), dz (shape=[n-1,m-1,l-1])
    """

    def integral(self, func: typing.Callable) -> ScalarType: raise NotImplementedError(f"")

    def average(self, func: typing.Callable) -> ScalarType: return self.integral(func)/self.measure

    def derivative(self,  *args, **kwargs): return NotImplemented

    def pullback(self, func,   *args, **kwargs):
        r"""
            ..math:: f:N\rightarrow M\\\Phi^{*}f:\mathbb{R}\rightarrow M\\\left(\Phi^{*}f\right)\left(u\right)&\equiv f\left(\Phi\left(u\right)\right)=f\left(r\left(u\right),z\left(u\right)\right)
        """

        return NotImplemented




import sys

from spdm.core.File import File
from ..util.logger import logger

if __name__ == '__main__':

    doc = File("/home/salmon/workspace/test_data/genray_profs_in.nc")

    logger.debug(doc.root["wz"])



from __future__ import annotations

import functools
import typing
from copy import copy
import numpy as np

from ..utils.logger import logger
from ..utils.typing import (ArrayLike, ArrayType, NumericType, array_type,
                            nTupleType)
from .GeoObject import GeoObject
from .PointSet import PointSet
from .Curve import Curve
from .Line import Segment


class Surface(PointSet):
    """ Surface """

    def __init__(self, *args, uv=None, **kwargs) -> None:
        super().__init__(*args, rank=2, **kwargs)
        self._edges: array_type = None
        self._cell: array_type = None
        self._uv = uv

    def enclose(self, *args) -> bool:
        if not self.is_closed:
            return False
        return super().enclose(*args)

    @property
    def boundary(self) -> Curve: return Curve(super().boundary, is_closed=True)

    @functools.cached_property
    def measure(self) -> float: return np.sum(self.dl)

    def coordinates(self, *uvw) -> ArrayType: raise NotImplementedError(f"{self.__class__.__name__}.coordinates")

    def derivative(self, *args, **kwargs): raise NotImplementedError(f"{self.__class__.__name__}.derivative")

    def remesh(self, u) -> Surface:
        other: Surface = copy(self)
        if isinstance(u, array_type):
            other._uv = u
        elif callable(u):
            other._uv = u(*self.points)
        else:
            raise TypeError(f"illegal type u={type(u)}")
        return other

    @property
    def edges(self) -> typing.Generator[Segment, None, None]:
        if self._points.ndim != 2:
            raise NotImplementedError(f"{self.__class__.__name__}.edges for ndim!=2")
        elif self._edges is None:
            for idx in range(self._points.shape[0]-1):
                yield Segment(self._points[idx], self._points[idx+1])
        elif isinstance(self._edges, array_type):
            for b, e in self._edges:
                yield Segment(self._points[b], self._points[e])
        else:
            raise NotImplementedError()



from __future__ import annotations

import typing
import numpy as np
import numpy.typing as np_tp
import functools
import collections

from copy import deepcopy

from ..utils.typing import ArrayType, array_type
from ..utils.numeric import float_nan, meshgrid, bitwise_and
from ..numlib.interpolate import interpolate

from .Functor import Functor
from .Path import update_tree


class DomainBase:
    """函数定义域"""

    _metadata = {"fill_value": float_nan}

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 1:
            if isinstance(args[0], dict):
                kwargs = collections.ChainMap(args[0], kwargs)
                args = args[1:]
            elif isinstance(args[0], (list,tuple)):
                args = args[0]

        if len(args) > 0:
            self._dims = args
        else:
            self._dims = kwargs.pop("dims", [])

        if len(kwargs) > 0:
            self._metadata = update_tree(deepcopy(self.__class__._metadata), kwargs)

    @property
    def label(self) -> str:
        return self._metadata.get("label", "unnamed")

    @property
    def is_simple(self) -> bool:
        return len(self._dims) > 0

    @property
    def is_empty(self) -> bool:
        return len(self._dims) == 0 or any([d == 0 for d in self._dims])

    @property
    def is_full(self) -> bool:
        return all([d is None for d in self._dims])

    @property
    def dims(self) -> typing.Tuple[ArrayType]:
        """函数的网格，即定义域的网格"""
        return self._dims

    @property
    def ndims(self) -> int:
        return len(self._dims)

    @property
    def shape(self) -> typing.Tuple[int]:
        return tuple([len(d) for d in self.dims])

    @functools.cached_property
    def points(self) -> typing.Tuple[ArrayType]:
        if len(self.dims) == 1:
            return self.dims
        else:
            return meshgrid(*self.dims, indexing="ij")

    @functools.cached_property
    def bbox(self) -> typing.Tuple[typing.List[float], typing.List[float]]:
        """函数的定义域"""
        return tuple(([d[0], d[-1]] if not isinstance(d, float) else [d, d]) for d in self.dims)

    @property
    def periods(self):
        return None

    def mask(self, *args) -> bool | np_tp.NDArray[np.bool_]:
        # or self._metadata.get("extrapolate", 0) != 1:
        if self.dims is None or len(self.dims) == 0 or self._metadata.get("extrapolate", 0) != "raise":
            return True

        if len(args) != self.ndim:
            raise RuntimeError(f"len(args) != len(self.dims) {len(args)}!={len(self.dims)}")

        v = []
        for i, (xmin, xmax) in enumerate(self.bbox):
            v.append((args[i] >= xmin) & (args[i] <= xmax))

        return bitwise_and.reduce(v)

    def check(self, *x) -> bool | np_tp.NDArray[np.bool_]:
        """当坐标在定义域内时返回 True，否则返回 False"""

        d = [child.__check_domain__(*x) for child in self._children if hasattr(child, "__domain__")]

        if isinstance(self._func, Functor):
            d += [self._func.__domain__(*x)]

        d = [v for v in d if (v is not None and v is not True)]

        if len(d) > 0:
            return np.bitwise_and.reduce(d)
        else:
            return True

    def eval(self, func, *xargs, **kwargs):
        """根据 __domain__ 函数的返回值，对输入坐标进行筛选"""

        mask = self.__domain__().mask(*xargs)

        mask_size = mask.size if isinstance(mask, array_type) else 1
        masked_num = np.sum(mask)

        if not isinstance(mask, array_type) and not isinstance(mask, (bool, np.bool_)):
            raise RuntimeError(f"Illegal mask {mask} {type(mask)}")
        elif masked_num == 0:
            raise RuntimeError(f"Out of domain! {self} {xargs} ")

        if masked_num < mask_size:
            xargs = tuple(
                [
                    (
                        arg[mask]
                        if isinstance(mask, array_type) and isinstance(arg, array_type) and arg.ndim > 0
                        else arg
                    )
                    for arg in xargs
                ]
            )
        else:
            mask = None

        value = func._eval(*xargs, **kwargs)

        if masked_num < mask_size:
            res = value
        elif is_scalar(value):
            res = np.full_like(mask, value, dtype=self._type_hint())
        elif isinstance(value, array_type) and value.shape == mask.shape:
            res = value
        elif value is None:
            res = None
        else:
            res = np.full_like(mask, self.fill_value, dtype=self._type_hint())
            res[mask] = value
        return res

    def interpolate(self, y: array_type, *args, **kwargs):
        x = self.points

        periods = self._metadata.get("periods", None)
        extrapolate = self._metadata.get("extrapolate", 0)

        return interpolate(*x, y, periods=periods, extrapolate=extrapolate)



import unittest

import numpy as np
from scipy import constants
from spdm.utils.logger import logger
from spdm.Mesh.Mesh import Mesh


class TestMesh(unittest.TestCase):

    def test_nullMesh(self):
        Mesh = Mesh()
        self.assertEqual(Mesh.type, None)
        self.assertEqual(Mesh.units, ["-"])
        self.assertEqual(Mesh.geometry, None)

    def test_structured_Mesh(self):
        from spdm.Mesh.StructuredMesh import StructuredMesh

        self.assertRaisesRegexp(
            TypeError, "Can't instantiate abstract class StructuredMesh with abstract method geometry", StructuredMesh, [10, 10])

    def test_uniform_mesh(self):
        Mesh = Mesh("uniform")
        self.assertEqual(Mesh.type, "uniform")
        self.assertEqual(Mesh.units, ["-"])
        self.assertEqual(Mesh.geometry, None)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
import collections
import collections.abc
import pathlib
import typing

import h5py
import numpy
from spdm.core.Entry import Entry
from spdm.core.File import File
from spdm.core.Path import Path
from spdm.utils.logger import logger
from spdm.utils.tags import _undefined_

SPDM_LIGHTDATA_MAX_LENGTH = 3


def h5_require_group(grp, path):
    if isinstance(path, str):
        path = path.split("/")

    for p in path:
        if isinstance(p, str):
            pass
        elif isinstance(p, int):
            if p < 0:
                num = len(grp)
                p = p % num
            p = f"__index__{p}"
        if grp is not None:
            grp = grp.require_group(p)
        else:
            raise KeyError(f"Cannot create group for {p}")

    return grp


def h5_put_value(grp, path, value):
    res = None
    if path is None:
        path = []
    elif isinstance(path, Path):
        path = path[:]
    elif not isinstance(path, list):
        path = [path]

    if isinstance(value, collections.abc.Mapping):
        grp = h5_require_group(grp, path)
        for k, v in value.items():
            h5_put_value(grp, [k], v)
    elif len(path) == 0:
        raise KeyError(f"Empty path!")
    else:
        grp = h5_require_group(grp, path[:-1])
        path = path[-1]

        if isinstance(path, int):
            path = f"__index__{path}"
        # elif not isinstance(path, str):
        #     raise KeyError(path)
        if path != "" and path in grp.keys():
            del grp[path]

        if type(value) is list:
            array_value = numpy.array(value)

            if array_value.dtype.type is numpy.object_:
                grp = h5_require_group(grp, path)

                grp.attrs["__is_list__"] = True

                for idx, v in enumerate(value):
                    h5_put_value(grp, idx, v)

            elif array_value.dtype.type is numpy.unicode_:
                # h5py does not support unicode string.
                array_value = array_value.astype(h5py.special_dtype(vlen=str))
                h5_put_value(grp, path, array_value)
            else:
                h5_put_value(grp, path, array_value)
        elif type(value) is numpy.ndarray and len(value) > SPDM_LIGHTDATA_MAX_LENGTH:
            grp[path] = value
        else:  # type(value) in [str, int, float]:
            grp.attrs[path] = value

    return res


def h5_get_value(obj, path=None, projection=None, default=_undefined_, **kwargs):
    if path is None:
        path = []
    elif isinstance(path, Path):
        path = path[:]
    elif not isinstance(path, list):
        path = [path]

    if obj is None:
        raise RuntimeError("None group")

    prefix = []
    if path is not None:
        for p in path:
            if isinstance(p, str):
                pass
            elif isinstance(p, int):
                if p < 0:
                    num = len(grp)
                    p = p % num
                p = f"__index__{p}"

            prefix.append(p)

            if p in obj:
                obj = obj[p]
            elif p in obj.attrs:
                obj = obj.attrs[p]
            else:
                raise KeyError(f"Can not search element at {'/'.join(prefix)} !")

    if projection is None:
        if isinstance(obj, h5py.Group):
            if obj.attrs.get("__is_list__", False):
                res = [h5_get_value(obj[k]) for k in obj]
            else:
                res = {**(h5_get_value(obj.attrs)), **{k: h5_get_value(obj[k]) for k in obj}}
        elif isinstance(obj, h5py.AttributeManager):
            res = {k: h5_get_value(obj[k]) for k in obj if not k.startswith("__")}
        elif isinstance(obj, h5py.Dataset):
            res = obj[:]
        else:
            res = obj
    elif isinstance(projection, str):
        if isinstance(obj, h5py.Group):
            res = h5_get_value(obj.attrs, projection) or h5_get_value(obj.get(projection, None))
        elif isinstance(obj, h5py.AttributeManager):
            res = h5_get_value(obj.get(projection, None))

    elif isinstance(obj, h5py.Group):
        if obj.attrs.get("__is_list__", False):
            res = []
        else:
            res = {
                **h5_get_value(obj.attrs, projection),
                **{k: h5_get_value(obj[k]) for k, v in projection.items() if v > 0 and k in obj},
            }
    elif isinstance(obj, h5py.AttributeManager):
        res = {k: h5_get_value(obj[k]) for k, v in projection.items() if v > 0 and k in obj}
    elif isinstance(obj, h5py.Dataset):
        res = obj[:]
    else:
        res = obj

    return res


def h5_dump(grp):
    return h5_get_value(grp, [])


@File.register(["h5", "hdf5", "HDF5"])
class HDF5File(File):
    MOD_MAP = {
        File.Mode.read: "r",
        File.Mode.read | File.Mode.write: "r+",
        File.Mode.write: "w-",
        File.Mode.write | File.Mode.create: "w",
        File.Mode.read | File.Mode.write | File.Mode.create: "a",
    }

    """
        r       Readonly, file must exist (default)
        r+      Read/write, file must exist
        w       Create file, truncate if exists
        w- or x Create file, fail if exists
        a       Read/write if exists, create otherwise
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fid = None

    @property
    def mode_str(self) -> str:
        return HDF5File.MOD_MAP[self.mode]

    def open(self) -> File:
        if self.is_open:
            return self

        try:
            if self._fid is None:
                self._fid = h5py.File(self.path, mode=self.mode_str)
        except OSError as error:
            raise FileExistsError(f"Can not open file {self.path}! {error}")
        else:
            logger.debug(f"Open HDF5 File {self.path} mode={self.mode}")

        super().open()

        return self

    def close(self):
        if not self.is_open:
            return
        if self._fid is not None:
            self._fid.close()
        self._fid = None
        return super().close()

    @property
    def entry(self) -> Entry:
        return H5Entry(self.open())

    def read(self, lazy=True) -> Entry:
        return H5Entry(self.open())

    def write(self, *args, **kwargs):
        H5Entry(self.open()).insert(*args, **kwargs)


@Entry.register(["h5", "hdf5", "HDF5"])
class H5Entry(Entry):
    def __init__(self, uri: str | HDF5File, *args, **kwargs):
        super().__init__(None, *args, **kwargs)

        if isinstance(uri, str):
            self._file = HDF5File(uri)
        elif isinstance(uri, File):
            self._file = uri
        else:
            raise TypeError(f"cache must be HDF5File or str, but got {type(uri)}")

        self._data = self._file._fid

    def __copy_from__(self, other: H5Entry) -> Entry:
        super().__copy_from__(other)
        self._file = other._file
        return self

    @property
    def is_writable(self) -> bool:
        return "r" in self._data.is_writable

    def copy(self, other):
        if isinstance(other, Entry):
            other = other.entry.__real_value__()
        self.put(None, other)

    # def put(self, path, value, *args, **kwargs):
    #     return h5_put_value(self._cache, path, value)

    # def get(self, path=[], projection=None, *args, **kwargs):
    #     return h5_get_value(self._cache, path, projection=projection)

    def insert(self, value, *args, **kwargs):
        return h5_put_value(self._data, self._path, value, *args, **kwargs)

    def find(self, *args, **kwargs) -> typing.Any:
        return h5_get_value(self._data, self._path, *args, **kwargs)

    def dump(self):
        return h5_dump(self._data)

    def iter(self, path, *args, **kwargs):
        raise NotImplementedError()


# class HDF5Collection(FileCollection):
#     def __init__(self, uri, *args, **kwargs):
#         super().__init__(uri, *args,
#                          file_extension=".h5",
#                          file_factory=lambda *a, **k: H5File(*a, **k),
#                          ** kwargs)



from __future__ import annotations

import typing
from enum import Flag, auto

from ..utils.plugin import Pluggable
from ..utils.uri_utils import URITuple, uri_split
from .Entry import Entry


class Document(Pluggable):
    """
        Connection like object
    """
    _plugin_registry = {}
    _plugin_prefix = "spdm.plugins.data.plugin_"

    class Mode(Flag):
        read = auto()       # open for reading (default)
        write = auto()      # open for writing, truncating the file first
        create = auto()     # open for exclusive creation, failing if the file already exists
        append = read | write | create
        temporary = auto()  # is temporary
    """
        r       Readonly, file must exist (default)
        r+      Read/write, file must exist
        w       Create file, truncate if exists
        w- or x Create file, fail if exists
        a       Read/write if exists, create otherwise
    """
    MOD_MAP = {Mode.read: "r",
               Mode.read | Mode.write: "rw",
               Mode.write: "x",
               Mode.write | Mode.create: "w",
               Mode.read | Mode.write | Mode.create: "a",
               }
    INV_MOD_MAP = {"r": Mode.read,
                   "rw": Mode.read | Mode.write,
                   "x": Mode.write,
                   "w": Mode.write | Mode.create,
                   "a": Mode.read | Mode.write | Mode.create,
                   }

    class Status(Flag):
        opened = auto()
        closed = auto()

    def __init__(self, uri, *args, mode: typing.Any = Mode.read, **kwargs):
        """
         r       Readonly, file must exist (default)
         rw      Read/write, file must exist
         w       Create file, truncate if exists
         x       Create file, fail if exists
         a       Read/write if exists, create otherwise
        """
        if self.__class__ is Document:
            return Document.__dispatch_init__(None, self, uri, *args, **kwargs)

        self._url = uri_split(uri)
        self._mode = Document.INV_MOD_MAP[mode] if isinstance(mode, str) else mode
        self._is_open = False

    def __del__(self):
        if getattr(self, "_is_open", False):
            self.close()

    def __str__(self): return f"<{self.__class__.__name__}  {self.url} >"

    @property
    def url(self) -> URITuple: return self._url

    @property
    def path(self) -> typing.Any: return self.url.path 

    @property
    def mode(self) -> Mode: return self._mode

    # @property
    # def mode_str(self) -> str:
    #     return ''.join([(m.name[0]) for m in list(Connection.Mode) if m & self._mode])

    @property
    def is_readable(self) -> bool: return bool(self._mode & Document.Mode.read)

    @property
    def is_writable(self) -> bool: return bool(self._mode & Document.Mode.write)

    @property
    def is_creatable(self) -> bool: return bool(self._mode & Document.Mode.create)

    @property
    def is_temporary(self) -> bool: return bool(self._mode & Document.Mode.temporary)

    @property
    def is_open(self) -> bool: return self._is_open

    def open(self) -> Document:
        self._is_open = True
        return self

    def close(self) -> None:
        self._is_open = False


    def __enter__(self) -> Document: return self.open()

    def __exit__(self, exc_type, exc_value, traceback): return self.close()

    def read(self, lazy=False) -> Entry: raise NotImplementedError()

    def write(self, data=None, lazy=False, **kwargs) -> Entry: raise NotImplementedError()

    @property
    def entry(self) -> Entry: raise NotImplementedError()




import collections
import re

from .logger import logger
from .multimap import Multimap


class Alias:
    """ Multi-Mapping of path alias

       Example:
            >>> alias=Alias()
            >>> alias.append("http://a.b.c.d.com/schemas/draft-00","/some/local/dir/%PLACEHOLDER%.json")
            >>> alias.append("http://a.b.c.d.com/schemas/draft-00/flow","/other/local/dir/")
            >>> [ p for p in alias.get("http://a.b.c.d.com/schemas/draft-00/flow/Entry")]
            ["/some/local/dir/flow/Node.json",
            "/other/local/dir/Node",
            "http://a.b.c.d.com/schemas/draft-00/flow/Node"]

            TODO (salmon 2020.04.14): need regex match
    """

    def __init__(self, *args,  **kwarg):
        self._mmap = collections.deque()

    def _compile(self, s, t):

        if not isinstance(s, re.Pattern):
            if '*' in s:
                s = s.replace('*', "(?P<_path_>[^?#]*)")+"(#(?P<fragment>.*))?"

            if '(' not in s:
                re_s = re.compile(f"{s}(?P<_path_>[^?#]*)(#(?P<fragment>.*))?")
            else:
                re_s = re.compile(s)

        if '*' in t:
            t = t.replace('*', '{_path_}')
        elif '{' not in t:
            t = t+'{_path_}'

        return re_s, t

    def match(self, *keys):
        """ if key is re.Pattern and  match(s)
            or key is string and s.startswith(key)
        """

        for s in keys:
            if s is None:
                continue
            for pattern, target in self._mmap:
                if isinstance(pattern, re.Pattern):
                    m = pattern.match(s)
                elif isinstance(pattern, str):
                    m = re.match(pattern, s)
                else:
                    m = None
                    logger.warning(
                        f"Type of key is not string or re.Pattern! {pattern}")
                if m is not None:
                    yield target.format(*m.groups(), **m.groupdict())

    def append(self, source: str, target: str):
        return self._mmap.append((self._compile(source, target)))

    def prepend(self, source: str, target: str):
        return self._mmap.appendleft((self._compile(source, target)))

    def append_many(self, m):
        if m is None:
            pass
        elif isinstance(m, collections.abc.Mapping):
            self.append_many(m.items())
        elif isinstance(m, collections.abc.Sequence):
            for k, v in m:
                self.append(k, v)
        else:
            raise TypeError(f"Require list or map, not [{type(m)}]")

    def prepend_many(self, m):
        if m is None:
            pass
        elif isinstance(m, collections.abc.Mapping):
            self.prepend_many(m.items())
        elif isinstance(m, collections.abc.Sequence):
            for k, v in m:
                self.prepend(k, v)
        else:
            raise TypeError(f"Require list or map, not [{type(m)}]")

    def add(self, *args, **kwargs):
        return self.append(*args, **kwargs)

    def remove(self, range_start: str, range_end: str = None):
        raise NotImplementedError()
        # self._mmap.remove(range_start, range_end)



import unittest
from copy import copy, deepcopy

import numpy as np
from spdm.data.Dict import Dict
from spdm.data.List import List
from spdm.data.HTree import HTree
from spdm.data.Path import Path
from spdm.utils.logger import logger


class TestNodeList(unittest.TestCase):
    data = {
        "a": [
            "hello world {name}!",
            "hello world2 {name}!",
            1, 2, 3, 4
        ],
        "c": "I'm {age}!",
        "d": {
            "e": "{name} is {age}",
            "f": "{address}"
        }
    }

    def test_insert(self):
        cache = {}

        d = Dict(cache)

        d["a"] = "hello world {name}!"
        d["c"].append(1.23455)
        d["c"].append({"a": "hello world", "b": 3.141567})
        self.assertEqual(cache["a"], "hello world {name}!")
        self.assertEqual(cache["c"][0],  1.23455)
        self.assertEqual(cache["c"][1]["a"], "hello world")
        self.assertEqual(d["c"][1]["a"].value, cache["c"][1]["a"])

        d["e"]["f"] = 5
        d["e"]["g"] = 6
        self.assertEqual(cache["e"]["f"], 5)
        self.assertEqual(cache["e"]["g"], 6)

    def test_find_by_cond(self):
        cache = [
            {"name": "wang wu", "age": 21},
            {"name": "wang liu", "age": 22},
            {"name": "li si",    "age": 22},
            {"name": "zhang san", "age": 24},
        ]

        d0 = List(cache)
        self.assertEqual(d0[({"name": "li si"}), "age"].value, 22)

        d1 = Dict({"person": cache})

        young = d1["person",  {"age": 22}]

        self.assertEqual(young[0, "name"].value,  "wang liu")
        self.assertEqual(young[1, "name"].value,  "li si")

        # res = d1["person", {"age": 22}]
        # names = [d["name"] for d in res]
        # self.assertEqual(len(names), 2)

    def test_insert_by_cond(self):
        cache = [
            {"name": "wang wu",   "age": 21},
            {"name": "wang liu",  "age": 22},
            {"name": "li si",     "age": 22},
            {"name": "zhang san", "age": 24},
        ]

        d0 = List(cache)

        d0[{"name": "wang wu"}, "address"] = "hefei"

        self.assertEqual(cache[0]["address"],  "hefei")

    def test_iter(self):
        d = List(self.data["a"])

        expected = [v for v in d]
        self.assertEqual(self.data["a"], expected)


if __name__ == '__main__':
    unittest.main()



import collections.abc
import typing
import uuid

import numpy as np
from spdm.core.HTree import HTree
from spdm.geometry.BBox import BBox
from spdm.geometry.Circle import Circle
from spdm.geometry.Curve import Curve
from spdm.geometry.GeoObject import GeoObject
from spdm.geometry.Line import Line
from spdm.geometry.Point import Point
from spdm.geometry.PointSet import PointSet
from spdm.utils.logger import logger

from spdm.view.View import View

SVG_TEMPLATE = """<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg xmlns:xlink="http://www.w3.org/1999/xlink" width="{p_width}pt" height="{p_height}pt" viewBox="{xmin:.2f} {ymin:.2f} {width:.2f} {height:.2f}"  xmlns="http://www.w3.org/2000/svg" version="1.1">
<style>
    .sp_geo_object {{
        stroke: {color}; 
        stroke-linejoin: miter; 
        stroke-width:{line_width};
    }}
</style>
    <g id="{name}">
    {contents}
    </g>
</svg>"""


@View.register(["svg", "SVG"])
class SVGView(View):

    TEMPLATE = SVG_TEMPLATE

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def draw(self, obj, **kwargs) -> typing.Any:

        bbox = []
        contents = []

        logger.debug(obj)

        if isinstance(obj, tuple):
            obj, opts = obj
        else:
            opts = {}

        if hasattr(obj.__class__, "__geometry__"):
            geo, *_ = obj.__geometry__(view=self.viewpoint)
            contents.append(geo)
            bbox.append(geo.bbox)
        elif hasattr(obj, "bbox"):
            contents.append(obj.bbox.__repr_svg__(**opts))
            bbox.append(obj.bbox)
        else:
            logger.error(f"Can not plot {obj}")

        if bbox is None or len(bbox) == 0:
            raise RuntimeError(f"SVGView.show() requires a bbox attribute ")

        bbox = np.bitwise_or.reduce(bbox)

        (xmin, ymin), (xmax, ymax) = bbox
        width = xmax-xmin
        height = ymax-ymin
        p_width = 300
        p_height = int(p_width*height/width+10)
        contents = "\n\t".join(contents)

        padd_x = (width/50)
        padd_y = (height/50)

        line_width = min(width/100, height/100)

        return self.TEMPLATE.format(
            p_width=p_width,
            p_height=p_height,
            xmin=xmin-padd_x,
            ymin=ymin-padd_y,
            width=width+2*padd_x,
            height=height+2*padd_y,
            contents=contents,
            signature=self.signature,
            name=kwargs.pop("name", f"spdm_"),
            title=kwargs.pop("title", ""),
            xlabel=kwargs.pop("xlabel", ""),
            ylabel=kwargs.pop("ylabel", ""),
            line_width=line_width,
            color="black",)

    def _draw_geo(self, cancas, obj: GeoObject | HTree | BBox, styles=None, **kwargs) -> str:

        name = kwargs.pop('name', obj.name)

        if getattr(obj, "ndim", 0) != 2:
            raise NotImplementedError(f"{self.__class__.__name__}.draw ndim={obj.ndim}")

        elif isinstance(obj, BBox):
            xmin = obj._xmin
            xmax = obj._xmax

            if np.allclose(obj._xmin, obj._xmax):
                svg = f'<line  x1="{xmin[0]}" y1="{xmin[1]}" x2="{xmin[0]}" y2="{xmin[1]}" class="sp_geo_object" />'
            else:
                svg = f'<rect x="{xmin[0]}" y="{xmin[1]}" width="{xmax[0]-xmin[0]}" height="{xmax[1]-xmin[1]}" class="sp_geo_object_bbox" />'

        elif not isinstance(obj, GeoObject):
            svg = f"<text x='0' y='0' fill='red'>{obj}</text>"

        elif isinstance(obj, Point):
            svg = f'<point cx="{obj.x}" cy="{obj.y}" r="0.01" class="sp_geo_object" />'

        elif isinstance(obj, PointSet):
            fill_or_not = "fill='none'" if obj.rank == 1 else ""

            return f'<polyline class="sp_geo_object" points="{" ".join([f"{x},{y}" for x, y in obj._points])}" {fill_or_not} />'

        elif isinstance(obj, Curve):

            pts = "M "
            pts += '\nL'.join([f' {x} {y}' for x, y in obj._points])
            if obj.is_closed:
                pts += " Z"
            svg = f'<path class="sp_geo_object" d="{pts}" fill="none" />'

        elif isinstance(obj, Line):
            x0, y0 = obj.p0
            x1, y1 = obj.p1
            svg = f'<line  x1="{xmin[0]}" y1="{xmin[1]}" x2="{xmin[0]}" y2="{xmin[1]}" class="sp_geo_object" />'
        elif isinstance(obj, Circle):
            svg = f'<circle cx="{obj.x}" cy="{obj.y}" r="{obj.r}" class="sp_geo_object" />'

        elif isinstance(obj, GeoObject):
            svg = self.draw(obj.bbox, name=f"< {obj.__class__.__name__} {name}/>", **kwargs)

        else:
            raise NotImplementedError(f"Draw {obj.__class__.__name__}")
        return svg


__SP_EXPORT__ = SVGView
